/*
   calcdist.c   - calculate a distance matrix for rows or columns of a character-sample matrix
   The input is a rectangular matrix of m rows and n columns,
   possibly split among several files, each having m lines.
   Optionally:
   - column names can occur in the first line of each file
   - row names can occur as the first word of each line
   - the output matrix can include appropriate names
   - the output can be in the form of a table
   Each line in a file contains data for all (if 1 file)
   or a portion of (if >1 files) a row in the matrix.
   Each line in a file must contain the same number of
   items.

   There is a choice of several distance measures 


 */

#define VERSION_STRING "CALCDIST V0.1"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#define USE_DOUBLE		/* #define USE_DOUBLE if you want to use doubles */

#ifdef USE_DOUBLE

typedef double Float;
#define SCANF_FLOAT_FORMAT "%lf"

#else

typedef float Float;
#define SCANF_FLOAT_FORMAT "%f"

#endif /* USE_DOUBLE */

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE (-1)
#endif
#ifndef NULL
#define NULL ((void *)0)
#endif

/* two definitions, not side-effect proof! */
#ifndef MIN
#define MIN(A,B) ((A<=B) ? (A) : (B))
#endif
#ifndef MAX
#define MAX(A,B) ((A>=B) ? (A) : (B))
#endif

#define INFILE_NAME	"infile"
#define OUTFILE_NAME "outfile"
#define TREEFILE_NAME "treefile"

/* options for the distance measure */

#define  EUCLIDEAN 				0
#define  BRAY_CURTIS 			1
#define  CANBERRA 				2
#define  JACCARD 				3
#define  SORENSEN 				4
#define  SIMPLE_MATCHING 		5
#define  BARONI_URBANI_BUSER	6
#define  MEAN_CENSORED_EUCLIDEAN 7
#define	 MANHATTAN				8
#define	NUM_DIST_MEASURES		9

/* these strings must appear in the same order as the constants above */

static char *distNames[] =
{
  "Euclidean Distance",
  "Bray-Curtis Distance",
  "Canberra Distance",
  "Jaccard Similarity",
  "Sorensen Similarity",
  "Simple Matching Coefficient",
  "Baroni-Urbani Buser Coefficient",
  "Mean Censored Euclidean Distance",
  "Manhattan Distance"
};

void 
getCounts (Float * v1, Float * v2, int length, int *a, int *b, int *c, int *d)
{
  int i;
  int counts[4];
  counts[0] = counts[1] = counts[2] = counts[3] = 0;

  for (i = 0; i < length; ++i)
    ++counts[((v1[i] > 0) ? 0 : 2) + ((v2[i] > 0) ? 0 : 1)];

  *a = counts[0];
  *b = counts[1];
  *c = counts[2];
  *d = counts[3];
}

double 
distJaccard (Float * v1, Float * v2, int length)
{
  int a, b, c, d, den;
  getCounts (v1, v2, length, &a, &b, &c, &d);

  den = a + b + c;

  /* return 1.0 - the Jaccard similarity: */

  return den ? (b + c) / (double) den : 0;
}

double 
distSorensen (Float * v1, Float * v2, int length)
{
  int a, b, c, d, den;
  getCounts (v1, v2, length, &a, &b, &c, &d);

  den = 2 * a + b + c;

  /* return 1.0 - the Sorensen similarity: */

  return den ? (b + c) / (double) den : 0;
}


double 
distSimpleMatching (Float * v1, Float * v2, int length)
{
  int a, b, c, d, den;
  getCounts (v1, v2, length, &a, &b, &c, &d);

  /* return 1.0 - the simple matching coefficient */

  den = a + b + c + d;

  return den ? (b + c) / (double) den : 0;
}

double 
distBaroniUrbani (Float * v1, Float * v2, int length)
{
  int a, b, c, d;
  double den;
  getCounts (v1, v2, length, &a, &b, &c, &d);

  /* return 1.0 - the Baroni-Urbani and Buser similarity: */

  den = sqrt (a * d) + a + b + c;

  return den ? (b + c) / den : 0;
}

double 
distEuclidean (Float * v1, Float * v2, int length)
{
  double dist = 0.0;
  double f;
  int i;

  for (i = 0; i < length; ++i)
    {
      f = v1[i] - v2[i];
      dist += f * f;
    }
  return sqrt (dist);
}

double 
distMeanCensoredEuclidean (Float * v1, Float * v2, int length)
{
  double dist = 0.0;
  double f;
  int i;
  int n = 0;

  for (i = 0; i < length; ++i)
    {
      if (v1[i] != 0 || v2[i] != 0)
	{
	  ++n;
	  f = v1[i] - v2[i];
	  dist += f * f;
	}
    }
  return (n > 0) ? sqrt (dist / n) : 0;
}

double 
distBrayCurtis (Float * v1, Float * v2, int length)
{
  double num = 0.0, denom = 0.0;
  int i;
  for (i = 0; i < length; ++i)
    {
      if (v1[i] >= v2[i])	/* inline the absolute value computation */
	num += v1[i] - v2[i];
      else
	num += v2[i] - v1[i];

      denom += v1[i] + v2[i];
    }
  return denom ? num / denom : 0;
}

double 
distCanberra (Float * v1, Float * v2, int length)
{
  double can = 0.0;
  double num;
  int i;
  for (i = 0; i < length; ++i)
    {
      if (v1[i] > 0 || v2[i] > 0)
	{
	  if (v1[i] >= v2[i])	// inline the absolute value computation

	    num = v1[i] - v2[i];
	  else
	    num = v2[i] - v1[i];

	  can += num / (v1[i] + v2[i]);
	}
    }
  return can / length;
}

double 
distManhattan (Float * v1, Float * v2, int length)
{
  double dist = 0.0;
  double f;
  int i;

  for (i = 0; i < length; ++i)
    {
      f = v1[i] - v2[i];
	  if (f <= 0)
	      dist -= f;
	  else
		  dist += f;
    }
  return dist;
}


/* define the type representing a function calculating distances between two
   vectors */

typedef double (distMeas) (Float *, Float *, int);

/* this is the array of pointers to distance functions; the order must
   match that of the constants above */

static distMeas *measureFuns[] =
{
  distEuclidean,
  distBrayCurtis,
  distCanberra,
  distJaccard,
  distSorensen,
  distSimpleMatching,
  distBaroniUrbani,
  distMeanCensoredEuclidean,
  distManhattan


};

/* options for the output format */

#define SQUARE_MATRIX	0
#define LOWER_T_MATRIX	1
#define UPPER_T_MATRIX	2
#define DISTANCE_TABLE	3

#define NUM_OUTPUT_OPTIONS 4

/* these strings must appear in the same order as the constants above */
static char *outputOptions[] =
{
  "square distance matrix",
  "lower triangular matrix",
  "upper triangular matrix",
  "table of distances"
};

/* strings representing the transformation option */

#define TRANS_NONE 0
#define TRANS_SQRT 1
#define TRANS_NORM 2
#define TRANS_STD  3

#define NUM_TRANS_OPTIONS 4

static char *transOptions[] =
{
	"none",
	"square root",
	"normalize",
	"standardize"
};

/* variables used throughout this program */

/* program options */

static int distMeasure;		/* distance measure */
static int transOption;		/* transformation type */
static int haveRowNames;	/* row names are present */
static int haveColNames;	/* column names are present */
static int outputFormat;	/* format of output */
static int numRows;		/* number of rows */
static int numCols;		/* number of columns */
static int numItems;		/* number of items between which distances are computed; either numRows or numCols */
static int numAttr;		/* number of attributes; i.e. length of vector representing a row or column */
static int calcByRows;		/* compute distances between rows, not columns */
static int numInfiles;		/* number of input files */
static int outputNames;		/* do we output names? */
static int tabDelimited;	/* are cells always delimited by tabs, allowing for empty ones? */
static int forcedBlanks;	/* number of implicit blank items to return when trailing TABs are missing */
static char **infileName;	/* name(s) of input file(s) */
static char *outfileName;	/* name of output file */
static FILE **infile;		/* pointer(s) to input file(s) */
static FILE *outfile;		/* pointer to output file */
static int  *numFileCols;	/* keep track of number of columns in each file */
static int  fileNum = 0;	/* keep track of which file we are currently reading from */
static int  colNum = 0;		/* keep track of which column we are reading */
static int	rowNum = 0;		/* same for row */
static char **rowNames;		/* names of rows */
static char **colNames;		/* names of columns */
static char **itemNames;	/* names of rows or columns, whichever are the items */
static char *numFmtString;	/* format string for numbers, defines precision to which they're output */
static char *nameFmtString;	/* format string for OTU names */

/* variables used in computing the distances */

static Float **inputMat;	/* input matrix */

/* miscellaneous */

/* these two must match: */

#define ITEM_BUFF_SIZE 150
#define ITEM_BUFF_SIZE_STRING "150"

char itemBuff[ITEM_BUFF_SIZE + 2];	/* buffer to hold text for an item */

void *
myCalloc (int number, int size)
{
  void *res = NULL;

  if (number && size)		/* don't try to allocate 0 bytes - some systems crap out */
    if (!(res = (void *) calloc (number, size)))
      printf ("Out of memory asking for %ld bytes.\n", size * (long) number);

  return res;
}

void 
freeMatrix (void **t, int num)
{
  /*      free the matrix pointed to by t */

  int i;

  if (t)
    {
      for (i = 0; i < num; ++i)
	if (t[i])
	  free (t[i]);
      free (t);
    }
}

#if(0)
void **
allocLowerTriangle (int num, int size)
{
/* 
   allocate a lower triangular matrix with room for diagonal elements
   the matrix will have num rows and elements of the given size */

  void **t;
  int i;

  do
    {
      if (!(t = myCalloc (num, sizeof (void *))))
	  break;

      for (i = 0; i < num; ++i)
	if (!(t[i] = myCalloc (i + 1, size)))
	  break;

      if (i == num)
	return t;

    }
  while (FALSE);

  freeMatrix (t, num);

  return NULL;
}
#endif 
void **
allocRectangle (int rows, int cols, int size)
{
/* 
   allocate a lower triangular matrix with room for diagonal elements
   the matrix will have num rows and elements of the given size */

  void **t;
  int i;

  do
    {
      if (!(t = myCalloc (rows, sizeof (void *))))
	  break;

      for (i = 0; i < rows; ++i)
	if (!(t[i] = myCalloc (cols, size)))
	  break;

      if (i == rows)
	return t;

    }
  while (FALSE);

  freeMatrix (t, rows);

  return NULL;
}


char *
mystrdup (char *s)
{
  char *res = (char *) myCalloc (sizeof (char), strlen (s) + 1);
  if (res)
    strcpy (res, s);
  return res;
}

#define NUM_REQUIRED_ARGS 2

void 
showUsage (char *name)
{
  int i;

  printf ("Usage: %s ROWS COLS [options...][inputfile...]", name);
  puts ("   where:");
  puts ("ROWS        is the number of rows in the input matrix");
  puts ("COLS        is the number of columns in the input matrix");
  puts ("inputFile   is a list of 0 or more files to be joined horizontally");
  puts ("   and options are:");
  puts ("-h          print this summary of usage");
  puts ("-nr         names of rows are present as the first word in each line");
  puts ("-nc         names of columns are present as the first line in each file");
  puts ("-o outFile  put the matrix into outfile, instead of printing to the terminal");
  puts ("-p printFmt the printf-style output format string; defaults to '.3g'");
  puts ("-r          compute distances between pairs of rows, rather than columns");
  puts ("-s          strip names, don't output them in the distance matrix");
  puts ("-u numPref  a prefix for adding to item numbers when no names are available");
  puts ("            this can be a printf-style format string with a '%%d' to");
  puts ("            represent where to substitute the item number, e.g. 'Item#%%d'");
  puts ("-t          items are always separated by TABs; a missing item is treated as 0");
  puts ("-d measure  distance measure, an integer from:");

  for (i = 0; i < NUM_DIST_MEASURES; ++i)
    printf ("               %2d - %s\n", i, distNames[i]);

  puts ("-f format   format of output, an integer from:");

  for (i = 0; i < NUM_OUTPUT_OPTIONS; ++i)
    printf ("               %2d - %s\n", i, outputOptions[i]);

  puts ("-x transform   pre-computation transformation, an integer from:");

  for (i = 0; i < NUM_TRANS_OPTIONS; ++i)
    printf ("               %2d - %s\n", i, transOptions[i]);

  puts ("The default behaviour is to read a matrix from standard input,");
  puts ("leave the data untransformed, and");
  puts ("compute the Euclidean distance between pairs of columns,");
  puts ("and print out a lower triangular matrix of distances, including zeros.");
  puts ("If more than one input file is listed, they are pasted together,");
  puts ("so that the nth row of the input matrix is the nth line of");
  puts ("file1, followed by the nth line of file2, and so on.");

  puts ("\nThis is " VERSION_STRING " by John Brzustowski (jbrzusto@gpu.srv.ualberta.ca)");
  puts ("You are free to use, copy, and modify it without fee.");
}

int 
getOptions (int argc, char **argv)
{
  int i;
  char c;
  int crapOut = FALSE;
  int numReqGotten = 0;		/* number of required arguments obtained so far */
  char *restOfArg;
//      int precision;                  /* number of digits after the decimal on output */

  /* set default options */

  distMeasure = EUCLIDEAN;
  transOption = TRANS_NONE;
  haveRowNames = haveColNames = FALSE;
  outputFormat = LOWER_T_MATRIX;
  numRows = numCols = 0;
  calcByRows = FALSE;
  numInfiles = 0;
  outputNames = TRUE;
  tabDelimited = FALSE;
  numFmtString = "%.3g";	/* default output format for calculated numbers */
  nameFmtString = "Item_%d";	/* default output format for integers used as names of items */

  /* create enough room to hold all possible input files
     a bit wasteful, but big deal */

  infileName = (char **) myCalloc (argc, sizeof (char *));
  infile = (FILE **) myCalloc (argc, sizeof (FILE *));
  numFileCols = (int *) myCalloc (argc, sizeof (int));

  outfileName = "Standard Output";
  outfile = stdout;

  i = 1;

  while (!crapOut)
    {
      if (i >= argc)
	{
	  if (numInfiles == 0)	/* if no input files, use stdin */
	    {
	      ++numInfiles;
	      infile[0] = stdin;
	      infileName[0] = "Standard_Input";
	    }
	  if (calcByRows)
	    {
	      numItems = numRows;
	      numAttr = numCols;
	    }
	  else
	    {
	      numItems = numCols;
	      numAttr = numRows;
	    }

	  if (numReqGotten == NUM_REQUIRED_ARGS)
	    return TRUE;
	  printf ("%s: missing required arguments. Use the -h option for help.", argv[0]);
	  return FALSE;
	}

      if (argv[i][0] != '-')	/* not an option, either a filename or a required arg */
	{
	  switch (numReqGotten)
	    {
	    case 0:
	      if ((1 != sscanf (argv[i], "%d", &numRows))
		  ||
		  (numRows <= 0)
		)
		{
		  printf ("%s: number of rows must be > 0\n", argv[0]);
		  crapOut = TRUE;
		}
	      ++numReqGotten;
	      break;

	    case 1:
	      if ((1 != sscanf (argv[i], "%d", &numCols))
		  ||
		  (numCols <= 0)
		)
		{
		  printf ("%s: number of columns must be > 0\n", argv[0]);
		  crapOut = TRUE;
		}
	      ++numReqGotten;
	      break;

	    default:
	      infileName[numInfiles] = argv[i];
	      if (!(infile[numInfiles] = fopen (argv[i], "r")))
		{
		  printf ("Couldn't open file %s for input.\n", argv[i]);
		  crapOut = TRUE;
		}
	      ++numInfiles;
	      break;
	    }
	}
      else
	{
	  c = argv[i][1];

	  switch (c)
	    {
	      /* first handle options that take no argument */
	    case 'r':
	      calcByRows = TRUE;
	      break;

	    case 's':
	      outputNames = FALSE;
	      break;

	    case 't':
	      tabDelimited = TRUE;
	      break;

	    case 'h':
	      showUsage (argv[0]);
	      crapOut = TRUE;
	      break;

	    default:
	      /* deal with options that need an argument */


	      if (c == '\0' || argv[i][2] == '\0')
		{
		  /* in case the user left space after the option letter */
		  if (++i < argc)
		    restOfArg = argv[i];
		  else
		    restOfArg = NULL;
		}
	      else
		{
		  restOfArg = 2 + argv[i];
		}
	      switch (c)
		{
		case 'o':
		  if (restOfArg)
		    {
		      outfileName = restOfArg;
		      if (!(outfile = fopen (outfileName, "w")))
			{
			  printf ("Couldn't open file %s for output.\n", outfileName);
			  crapOut = TRUE;
			}
		    }
		  break;

		case 'd':
		  if ((1 != sscanf (restOfArg, "%d", &distMeasure))
		      ||
		      (distMeasure < 0 || distMeasure >= NUM_DIST_MEASURES)
		    )
		    {
		      printf ("%s: distance measure must be between 0 and %d; you specified %d\n", argv[0], NUM_DIST_MEASURES - 1, distMeasure);
		      crapOut = TRUE;
		    }
		  break;

		case 'x':
		  if ((1 != sscanf (restOfArg, "%d", &transOption))
		      ||
		      (transOption < 0 || transOption >= NUM_TRANS_OPTIONS)
		    )
		    {
		      printf ("%s: transformation option must be between 0 and %d; you specified %d\n", argv[0], NUM_TRANS_OPTIONS - 1, transOption);
		      crapOut = TRUE;
		    }
		  break;

		case 'p':	/* user format string for outputting numbers */
		  if (*restOfArg == '%')	/* skip the % if user has specified it */
		    ++restOfArg;
		  numFmtString = myCalloc (strlen (restOfArg + 2), sizeof (char));
		  sprintf (numFmtString, "%%%s", restOfArg);
		  break;

		case 'u':	/* user format string for outputting names */
		  if (!strchr (restOfArg, '%'))
		    {
		      nameFmtString = myCalloc (strlen (restOfArg) + 2 + 3, sizeof (char));
		      strcpy (nameFmtString, restOfArg);
		      strcat (nameFmtString, "%d");
		    }
		  else
		    {
		      nameFmtString = restOfArg;
		    }
		  break;

		case 'n':
		  switch (*restOfArg)
		    {
		    case 'c':
		    case 'C':
		      haveColNames = TRUE;
		      break;
		    case 'r':
		    case 'R':
		      haveRowNames = TRUE;
		      break;
		    default:
		      printf ("%s: argument to -%%c option must be 'r' or 'c'\n", restOfArg);
		      crapOut = TRUE;
		    };
		  break;

		case 'f':
		  if ((1 != sscanf (restOfArg, "%d", &outputFormat))
		      ||
		    (outputFormat < 0 || outputFormat >= NUM_OUTPUT_OPTIONS)
		    )
		    {
		      printf ("%s: output format (-f option) must be between 0 and %d\n", argv[0], NUM_OUTPUT_OPTIONS - 1);
		      crapOut = TRUE;
		    }
		  break;

		default:
		  printf ("%s: unknown option %s\n", argv[0], argv[i - 1]);
		  puts ("Use the -h option for help.");
		  return FALSE;
		}
	    }

	};

      ++i;			/* next arg */

    };
  return FALSE;			/* shouldn't get here */
}


int 
nextItem (int countCols)
{
  /* get the next item from the appropriate file 
     This means checking to see whether we've reached the end of line
     on the current file, and if so, moving to the next file
     An awkward function, but prevents having to allocate possibly very
     large line buffers that fgets would require.  We don't know the
     size we'd need, anyway.
   */

  char c;
  int i;
  static int notUsingTabsMsg = FALSE;

  if (tabDelimited)
    {
      for (;;)
	{
	  if (forcedBlanks > 0)
	  {
	   	itemBuff[0] = '\0';
		-- forcedBlanks;
		return TRUE;
      }

	  c = fgetc (infile[fileNum]);
	  if (c == EOF)
	    {
	      ++fileNum;
	      if (fileNum >= numInfiles)
		return FALSE;	/* done reading all files, no item found */
	      break;
	    }
	  else
	    {
	      for (i = 0; (isgraph (c) || c == ' ') && i < ITEM_BUFF_SIZE; ++i)
		{
		  if (c == ' ')
		    c = '_';

		  itemBuff[i] = c;
		  c = fgetc (infile[fileNum]);
		}
	      itemBuff[i] = '\0';
	      if (!isspace (c))	/* if the extra char is not just the expect TAB or EOL delimiter */
		ungetc (c, infile[fileNum]);	/* then return it */
	      if (c != '\t' && (!notUsingTabsMsg) && i == ITEM_BUFF_SIZE)
		{
		  puts ("Hmmm... this file appears to use spaces, not TABS to separate items, and\n"
			"I'm not smart enough to interpret a file with missing data and no TABS!");
		  notUsingTabsMsg = TRUE;
		  return FALSE;
		}

	      if (c == '\r' || c == '\n')
		{
		  c = fgetc (infile[fileNum]);	/* just in case it's a CR-LF pair, get the next char */
		  if (c != '\n')
		    ungetc (c, infile[fileNum]);
		  if (haveRowNames && tabDelimited && colNum < numFileCols[fileNum])
		  {
		   /* we've run out of columns - this file might be missing some trailing TABs */
		  	printf ("Warning: file %s is missing the last %d columns in row %d.\n"
					"I'm treating them as blanks.\n", infileName[fileNum], forcedBlanks = numFileCols[fileNum] - colNum - 1, rowNum);

		  }


		  if (countCols)
			 numFileCols [fileNum] = colNum;
		  ++rowNum;
		  ++fileNum;	/* end of line on this file, move to the next */
		  if (fileNum >= numInfiles)
		    fileNum = 0;
		  colNum = 0;
		}
        if (countCols)
			++colNum;
      	return TRUE;
	    }
	}
    }
  else
    {
      for (;;)
	{
	  c = fgetc (infile[fileNum]);
	  switch (c)
	    {
	    case EOF:
	      ++fileNum;
	      if (fileNum >= numInfiles)
		return FALSE;	/* done reading all files, no item found */
	      break;

	    case '\r':
	    case '\n':
	      c = fgetc (infile[fileNum]);	/* just in case it's a CR-LF pair, get the next char */
	      if (c != '\n')
		ungetc (c, infile[fileNum]);

	      ++fileNum;	/* end of line on this file, move to the next */
	      if (fileNum >= numInfiles)
		fileNum = 0;

	      break;

	    default:

	      if (c == '\t' || c == ' ')
		break;
	      for (i = 0; isgraph (c) && i < ITEM_BUFF_SIZE; ++i)
		{
		  itemBuff[i] = c;
		  c = fgetc (infile[fileNum]);
		}

	      ungetc (c, infile[fileNum]);	/* return extra character */

	      itemBuff[i] = '\0';
	      return TRUE;
	    }
	}
    }
  return FALSE;			/* shouldn't get here anyway */
}

int 
grabData ()			/* any input files are already open */
{
  int i, j;
  Float f;
  char *retPtr;

  i = j = 0;

  do
    {

      /* allocate space for the input matrix */

      if (calcByRows)
	{
	  if (!(inputMat = (Float **) allocRectangle (numRows, numCols, sizeof (Float))))
	    break;
	}
      else
	{
	  if (!(inputMat = (Float **) allocRectangle (numCols, numRows, sizeof (Float))))
	    break;
	}

      /* allocate a list of names, if there are any */

      if (haveRowNames && !(rowNames = (char **) myCalloc (numRows, sizeof (char *))))
	  break;

      if (haveColNames && !(colNames = (char **) myCalloc (numCols, sizeof (char *))))
	  break;

      /* read data from the files */

      if (haveColNames)
	{
	  /*

	     BUG FIX:  jmb 17 September '97
	     Throw away the top left corner entry, which should be blank, 
	     if items are tab delimited and row names are present.
	     This entry corresponds to the empty cell in
	     the top left corner of such a matrix.

	   */

	  if (tabDelimited & haveRowNames)
	    nextItem (0);

	  for (j = 0; j < numCols; ++j)
	    {
	      if (!nextItem (1))
		{
		  puts ("Couldn't read column name.");
		  goto crapOut;
		}

	      if (!(colNames[j] = mystrdup (itemBuff)))
		{
		  goto crapOut;
		}
	    }
	  if (j < numCols)
	    break;
	}

      for (i = 0; i < numRows; ++i)
	{
	  if (haveRowNames)
	    {
	      if (!nextItem (0))
		{
		  puts ("Couldn't read row name.");
		  goto crapOut;
		}
	      if (!(rowNames[i] = mystrdup (itemBuff)))
		{
		  goto crapOut;
		}
	    }
	  for (j = 0; j < numCols; ++j)
	    {
	      if (!nextItem (1))
		{
		  puts ("Ran out of data.");
		  goto crapOut;
		}
	      if (*itemBuff == '\0')	/* treat empty cell as zero */
		f = 0;
	      else
		{
		  f = strtod (itemBuff, &retPtr);
		  if (retPtr == itemBuff)
		    {
		      printf ("Looking for number but found '%s'.\n", itemBuff);
		      goto crapOut;
		    }
		}

	      if (calcByRows)
		inputMat[i][j] = f;
	      else
		inputMat[j][i] = f;
	    }
	}
      for (i = 0; i < numInfiles; ++i)
	fclose (infile[i]);

      if (calcByRows)
	itemNames = rowNames;
      else
	itemNames = colNames;

      return TRUE;

    crapOut:
      printf ("Error occurred at row %d, column %d.\n", i + 1, j + 1);
    }
  while (FALSE);

  return FALSE;
}


void 
outputName (int i)
{
/*
   output an item name
   If we are supposed to output names, use the name if available,
   otherwise format the item number according to a user string.
   If we are not outputting names, just print the item number.
   All numbers are origin 1.
 */

  if (outputNames)
    if (itemNames)
      fputs (itemNames[i], outfile);
    else
      fprintf (outfile, nameFmtString, i + 1);
  else
    fprintf (outfile, "%d", i + 1);
}

void 
TAB ()
{
  fputc ('\t', outfile);
}

void 
outputResults ()
{
  int i, j, jmin, jmax;
  int iLT, jLT;
  distMeas *dm = measureFuns[distMeasure];

  fprintf (outfile, "%d\n", numItems);
  if (outputFormat == DISTANCE_TABLE)
    {
      for (i = 1; i < numItems; ++i)
	{
	  for (j = 0; j < i; ++j)
	    {
	      outputName (i);
	      TAB ();
	      outputName (j);
	      TAB ();
	      fprintf (outfile, numFmtString, (*dm) (inputMat[i], inputMat[j], numAttr));
	      fputc ('\n', outfile);
	    }
	}
    }
  else
    {
      jmin = 0;
      jmax = numItems - 1;
      for (i = 0; i < numItems; ++i)
	{
	  if (outputFormat == LOWER_T_MATRIX)
	    jmax = i;
	  else if (outputFormat == UPPER_T_MATRIX)
	    jmin = i;

	  /* Here, we only output something for a name, even
	     if it is a number, when outputNames is true.
	     Unlike the distance table, where something
	     must be output to represent which item is
	     referred to, this distance matrix can be
	     interpreted without any labels at all */

	  if (outputNames)
	    {
	      outputName (i);
	      TAB ();
	    }

	  for (j = jmin; j <= jmax; ++j)
	    {
	      if (i >= j)
		{
		  iLT = i;
		  jLT = j;
		}
	      else
		{
		  iLT = j;
		  jLT = i;
		}
	      fprintf (outfile, numFmtString, (*dm) (inputMat[iLT], inputMat[jLT], numAttr));
	      if (j < jmax)
		fputc ('\t', outfile);
	    }
	  fputc ('\n', outfile);
	}
    }

}

void transformData ()
{
	int i, j;
	double sum, sumsq, mean, sd, mag;

	switch (transOption)
	{
		case TRANS_NONE:
			break;
		
		case TRANS_SQRT:
			for (i = 0; i < numRows; ++i)
				for (j = 0; j < numCols; ++j)
					if (inputMat[i][j] >= 0)
						inputMat[i][j] = sqrt (inputMat[i][j]);
					else
						inputMat[i][j] = - sqrt (-inputMat[i][j]);

			break;

		case TRANS_STD:
			if (calcByRows)
			{
				for (j=0; j < numCols; ++j)
				{
					sum = sumsq = 0;
					for (i=0; i < numRows; ++i)
					{
						sum += inputMat[i][j];
						sumsq += inputMat[i][j] * inputMat[i][j];
					}
					mean = sum / numRows;
					sd = sqrt ((sumsq - sum*sum/numRows)/(numRows-1));
					if (sumsq > 0)
						for (i=0; i < numRows; ++i)
							inputMat[i][j] = (inputMat[i][j] - mean) / sd;
				}
			}
			else
			{
				for (i=0; i < numRows; ++i)
				{
					sum = sumsq = 0;
					for (j=0; j < numCols; ++j)
					{
						sum += inputMat[i][j];
						sumsq += inputMat[i][j] * inputMat[i][j];
					}
					mean = sum / numCols;
					sd = sqrt ((sumsq - sum*sum/numCols)/(numCols-1));
					if (sumsq > 0)
						for (j=0; j < numCols; ++j)
							inputMat[i][j] = (inputMat[i][j] - mean) / sd;
				}
			}
			break;

		case TRANS_NORM:
			if (calcByRows)
			{
				for (j=0; j < numCols; ++j)
				{
					sumsq = 0;
					for (i=0; i < numRows; ++i)
						sumsq += inputMat[i][j] * inputMat[i][j];
					mag = sqrt (sumsq);
					if (mag > 0)
						for (i=0; i < numRows; ++i)
							inputMat[i][j] /= mag;
				}
			}
			else
			{
				for (i=0; i < numRows; ++i)
				{
					sumsq = 0;
					for (j=0; j < numCols; ++j)
						sumsq += inputMat[i][j] * inputMat[i][j];
					mag = sqrt (sumsq);
					if (mag > 0)
						for (j=0; j < numCols; ++j)
							inputMat[i][j] /= mag;
				}
			}
			break;
	}
}




int 
main (int argc, char **argv)
{
  do
    {
      if (!getOptions (argc, argv))
	break;

      if (!grabData ())
	break;
	  transformData ();
      outputResults ();		/* distances are now calculated when needed for output */
      exit(0);
    }
  while (FALSE);
  exit(1);
}
