/*
	qclust.c	- quickly cluster items via a SAHN method or Saitou/Nei neighbour-joining

	SAHN are Sequential Agglomerative Hierarchical Non-overlapping
	clustering algorithms. Saitou/Nei fits this, but has a more complicated
	rule for determining which pair of nodes is merged at any step.
	Also, the last interior node is ommitted in Saitou/Nei, so that
	all interior nodes have degree 3.	I treat this method separately.

    The O(n^2) (average?) SAHN algorithm is after the FORTRAN HC.F by Fionn Murtagh,
	available as package multiv from Statlib, with a small fix by me.

	The O(n^3) (max.) Neighbour code is by me, and is a slight optimization of
	the method described in

		J. Studier and K. Keppler (1988) Mol. Biol. Evol. 5(6):729-731
		
	Trees are output in sorted order, regardless of the clustering method.
	A tree is sorted when each of its subtrees is sorted, and
	the subtrees are ordered according to their smallest leaf nodes 
	(smallest to largest).

	The Phylip package of population genetics programs is by Joel Felsenstein et al.

	NEIGHBOR from Phylip uses the same output order, except that
	in the case of neighbour-joining, the ordering of the top-level
	subtrees follows a different rule.

	The convention for INFILE, OUTFILE, and TREEFILE is much as for NEIGHBOR,
	although command line options allow some flexibility.
	For details, invoke qclust with the '-h' option.
	
	This C program was translated from the Java embodied in
	a relatively easy to use web page calculator available at

		http://www.biology.ualberta.ca/jbrzusto/cluster.html

	It should be vanilla C.  It has compiled successfully under GCC
	and whatever cc is on IBM AIX.
			
	John Brzustowski	jbrzusto@gpu.srv.ualberta.ca
	14 June 1997

*/

#define VERSION_STRING "QCLUST V0.2"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define USE_DOUBLE	/* comment this out if you want to use doubles
						this can affect the topology of the tree, apparently */ 
//#define DO_TIMING	/* comment this out if you don't want the elapsed time displayed */

#ifdef DO_TIMING
#include <time.h>
#endif /* DO_TIMING */


#ifdef USE_DOUBLE

typedef double Float;
#define SCANF_FLOAT_FORMAT "%lf"

#else

typedef float Float;	
#define SCANF_FLOAT_FORMAT "%f"

#endif /* USE_DOUBLE */

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE (-1)
#endif
#ifndef NULL
#define NULL ((void *)0)
#endif

/* two definitions, not side-effect proof! */
#ifndef MIN
#define MIN(A,B) ((A<=B) ? (A) : (B))
#endif
#ifndef MAX
#define MAX(A,B) ((A>=B) ? (A) : (B))
#endif

#define INFILE_NAME	"infile"
#define OUTFILE_NAME "outfile"
#define TREEFILE_NAME "treefile"

/* options for the clustering method */

#define SINGLE_LINKAGE					0
#define COMPLETE_LINKAGE				1
#define UNWEIGHTED_ARITHMETIC_AVERAGE	2
#define WEIGHTED_ARITHMETIC_AVERAGE		3
#define UNWEIGHTED_CENTROID				4
#define WEIGHTED_CENTROID				5
#define WARDS_MINIMUM_VARIANCE			6
#define NEIGHBOUR_JOINING				7
#define NUM_CLUSTERING_METHODS			8

/* these strings must appear in the same order as the constants above */
static char *methodNames[] =
{
	"Single Linkage",
	"Complete Linkage",
	"Unweighted Arithmetic Average",
	"Weighted Arithmetic Average",
	"Unweighted Centroid",
	"Weighted Centroid",
	"Ward's Minimum Variance",
	"Saitou and Nei Neighbour Joining"
};

/* options concerning the location of item names*/

#define NAMES_BEFORE_DATA	0
#define NAMES_IN_DATA		1
#define NO_NAMES			2

#define NUM_NAME_OPTIONS	3

/* these strings must appear in the same order as the constants above */
static char *nameOptions[] =
{
	"all names after the item count but before the matrix",
	"each item's name before its list of distances to other items",
	"no names supplied - use numbers to identify items"
};

/* options for the distance matrix format */

#define SQUARE_MATRIX	0
#define LOWER_T_MATRIX	1
#define UPPER_T_MATRIX	2

#define NUM_MATRIX_OPTIONS 3

/* these strings must appear in the same order as the constants above */
static char *matrixOptions[] =
{
	"square distance matrix",
	"lower triangular matrix",
	"upper triangular matrix"
};


/* variables used throughout this program */

/* program options */

static int		clusterMethod;	/* clustering method */
static int		nameOption;		/* name option */
static int		matrixOption;	/* matrix format */
static int		compareDist;	/* do we compare the tree distances to the original distances */
static int		showProgress;	/* do we show progress on stdout */
static int		preserveOrder;/* only join adjacent clusters, thus preserving input node order */
static char*	infileName;		/* name of input file */
static char*	outfileName;	/* name of output file */
static char*	treefileName;	/* name of tree output file */
static FILE*	outfile;
static FILE*	treefile;

/* variables used in clustering */

static int		numItems = 0;	/* number of items */
static Float**	dist;			/* lower triangular distance matrix */
static Float**	distClone;		/* a place to save a copy of the original distances */
static char**	itemNames;		/* names of items */
static int*		clustNode1;		/* indices of first cluster joined at step i */
static int*		clustNode2;		/* indices of second cluster joined at step i */

static Float*	mergeDist1;		/* distance of node 1 to merger */
static Float*	mergeDist2;		/* distance of node 2 to merger (only used for Neighbour Joining) */

static int		numGroups;		/* number of groups into which the final tree should be collapsed */
										/* zero means don't collapse */

static int*		groupNums;		/* group numbers indexed by node numbers */

/* structure to hold comparison results, for lower stack overhead during recursion */

typedef struct compInfoStruct
{
	Float	sumDiffs;		/* sum of magnitude of differences between input and computed distances */
	Float	sumSqDiffs;		/* sum of squares of differences between input and computed distances */
	Float	maxDiff;		/* maximum difference beteen input and computed distances */
	int		whichMax1;		/* indexes of nodes with maximum difference */
	int		whichMax2;
	Float** treeDist;		/* array to build up tree distances in */
	Float** inputDist;		/* array of original distances */
	int*	deIndex;		/* array of integers to de-Index infix order */

} *compInfo;

/* miscellaneous */

/* these two must match: */

#define NAME_BUFF_SIZE 150
#define NAME_BUFF_SIZE_STRING "150"

char nameBuff[NAME_BUFF_SIZE+1];	/* buffer to grab a name */


void* myCalloc (int size, int number)
{
	void* res;
	
	if (! (res = (void*) calloc (size, number)))
		printf ("Out of memory asking for %ld bytes.\n", size * (long) number);

	return res;
}

void* myRealloc (void* p, int size)
{
	void* res;
	
	if (! (res = (void*) realloc (p, size)))
		printf ("Out of memory asking for %d bytes.\n", size);

	return res;
}

void freeMatrix (void** t, int num)
{
	/*	free the matrix pointed to by t */
	
	int i;

	if (t)
	{
		for (i=0; i < num; ++i)
			if (t[i])
				free (t[i]);
		free (t);
	}
}

void** allocLowerTriangle (int num, int size)
{
/* 
	allocate a lower triangular matrix with room for diagonal elements
	the matrix will have num rows and elements of the given size */

	void **t;
	int i;

	do
	{
		if (! (t = myCalloc (num, sizeof (void *))))
			break;

		for (i=0; i < num; ++i)
			if (! (t[i] = myCalloc (i+1, size)))
				break;

		if (i == num)
			return t;

	} while (FALSE);

	freeMatrix (t, num);

	return NULL;
}


char* mystrdup (char* s)
{
	char* res = (char *) myCalloc (sizeof (char), strlen (s) + 1);
	if (res)
		strcpy (res, s);
	return res;
}

int getOptions (int argc, char** argv)
{
	int i;
	char c;
	int crapOut = FALSE;
	char* restOfArg;

	/* set default options */	
	infileName = INFILE_NAME;
	outfileName = OUTFILE_NAME;
	treefileName = TREEFILE_NAME;
	showProgress = TRUE;
	compareDist = TRUE;
	clusterMethod = NEIGHBOUR_JOINING;
	nameOption = NAMES_IN_DATA;
	matrixOption = SQUARE_MATRIX;
	preserveOrder = FALSE;
   numGroups = 0;
			
	i = 1; 
	
	while (! crapOut)
	{
		if (i >= argc)
			return TRUE;

		if (argv[i][0] != '-')	/* not an option */
			break;			

		c = argv[i][1];


		switch (c)
		{
		/* first handle options that take no argument */
		case 'q':
			showProgress = FALSE;
			break;

		case 'h':
			crapOut = TRUE;
			break;

		case 'd':
			compareDist = FALSE;
			break;

		case 'p':
			preserveOrder = TRUE;
			break;

		default:
			/* deal with options that need an argument */
			

			if (c == '\0' || argv[i][2] == '\0')
			{
			/* in case the user left space after the option letter */
				if (++i < argc)
					restOfArg = argv [i];
				else
					restOfArg = NULL;
			}
			else
			{
				restOfArg = 2 + argv [i];
			}
			switch (c)
			{
         case 'g':
				if ( (1 != sscanf (restOfArg, "%d", &numGroups))
					||
					 (numGroups < 0)
					)
					crapOut = TRUE;
            break;

			case 'i':
				if (restOfArg)
					infileName = restOfArg;
				break;

			case 'o':
				if (restOfArg)
					outfileName = restOfArg;
				break;

			case 't':
				if (restOfArg)
					treefileName = restOfArg;
				break;

			case 'c':
				if ( (1 != sscanf (restOfArg, "%d", &clusterMethod))
					||
					 (clusterMethod < 0	|| clusterMethod >= NUM_CLUSTERING_METHODS)
					)
					crapOut = TRUE;

				if (clusterMethod != NEIGHBOUR_JOINING)
					compareDist = FALSE;				/* don't do RMS error calculation for other clustering methods */

				break;

			case 'n':
				if ( (1 != sscanf (restOfArg, "%d", &nameOption))
					||
					 (nameOption    < 0	|| nameOption	 >= NUM_NAME_OPTIONS)
					)
					crapOut = TRUE;
				break;

			case 'm':
				if ( (1 != sscanf (restOfArg, "%d", &matrixOption))
					||
					 (matrixOption  < 0	|| matrixOption  >= NUM_MATRIX_OPTIONS)
					)
					crapOut = TRUE;
				break;

			default:
				printf ("%s: unknown option %s\n", argv[0], argv[i-1]);
				puts ("Use the -h option for help.");
				return FALSE;
				break;
			}

		};

		++i;	/* next arg */

	}

	printf	("Usage: %s [options...]", argv[0]);
	puts	("   where options are:");
	puts	("-h          print this summary of usage");
	puts	("-q          quiet: don't show progress on terminal");
	puts	("-d          for Neighbour Joining method: don't compute RMS error");
	puts  ("-p          preserve order: only adjacent clusters are ever joined");
	puts  ("                (useful for stratigraphic data)");
	puts	("-i inFile   get the item count and distance matrix from inFile");
	puts	("-o outFile  put the tree topology and lengths into outFile");
	puts	("-t treeFile put the Phylip DRAWTREE-style tree into treeFile");
   puts	("-g numGrp   group items into numGrp groups by first clustering, then");
   puts	("            collapsing the tree bottom-up until only numGrp leaf nodes");
   puts  ("            remain.  The items in each collapsed subtree are listed.");
	puts	("-c method   cluster method, an integer from:");
	
	for (i=0; i < NUM_CLUSTERING_METHODS; ++i)
		printf ("               %2d - %s\n", i, methodNames [i]);


	puts	("-n nameLoc  location of item names, an integer from:");

	for (i=0; i < NUM_NAME_OPTIONS; ++i)
		printf ("               %2d - %s\n", i, nameOptions [i]);

	puts	("-m matFmt    matrix format, an integer from:\n");

	for (i=0; i < NUM_MATRIX_OPTIONS; ++i)
		printf ("               %2d - %s\n", i, matrixOptions [i]);

	puts	("The default behaviour is to perform Neighbour-joining");
	puts	("on the square matrix stored in 'infile'. The first word");
	puts	("of infile is the number of items.  For each item,");
	puts	("its name and then its distances to other items, including itself,");
	puts	("are read from infile.  The progress of clustering is shown.");
	puts	("The textual tree picture and legths are put in 'outfile', and");
	puts	("the Phylip DRAWTREE string is put in 'treefile'.");

	puts	("\nThis is " VERSION_STRING " by John Brzustowski (jbrzusto@gpu.srv.ualberta.ca)");
	puts	("You are free to use, copy, and modify it without fee.");
	puts	("There is absolutely no warranty.");
	return FALSE;
}

FILE *myfopen (char *name, char *mode)
{
	int stdFlag;
	stdFlag = (!name[0]) || (name[0] == '-' && name[1] == '\0');
	if (stdFlag)
	{
		if (*mode == 'r')
			return stdin;
		else
			return stdout;
	}
	return fopen (name, mode);
}

void myfclose (FILE *f)
{
	if (f != stdin && f != stdout)
		fclose (f);
}

int grabData (char* ifName, int saveCopy)
{
	FILE *fi;
	int i, j;
	int firstColumn;
	Float f;

	do
	{
		fi = myfopen (ifName, "r");
		if (!fi)
		{
			printf ("Couldn't open file %s for reading.\n", ifName);
			break;
		}

		if (EOF == fscanf (fi, "%d", & numItems))
		{
			printf ("Couldn't read the number of items from file %s.\n", ifName);
			break;
		}

		if (numItems < 2)
		{
			printf ("Need numItems >= 2, but %s has numitems = %d\n", ifName, numItems);
			break;
		}

	/* allocate space for names, distances, intermediates and results */
		
		/* allocate a lower triangular Float matrix, including the diagonal */

		if ( ! (dist = (Float**) allocLowerTriangle (numItems, sizeof (Float))))
			break;
		
		/* allocate a list of names, if there are any */

		if (nameOption == NO_NAMES)
			itemNames = NULL;
		else if ( ! (itemNames = (char**) myCalloc (sizeof (char*), numItems)))
			 break;

		if ( ! (clustNode1 = (int*) myCalloc (sizeof (int), numItems-1)))
			break;

		if ( ! (clustNode2 = (int*) myCalloc (sizeof (int), numItems-1)))
			break;

		if ( ! (mergeDist1 = (Float*) myCalloc (sizeof (Float), numItems-1)))
			break;

		if (clusterMethod == NEIGHBOUR_JOINING)
			if ( ! (mergeDist2 = (Float*) myCalloc (sizeof (Float), numItems-1)))
				break;

		/* read data from the file */
		
		if (nameOption == NAMES_BEFORE_DATA)
		{
			for (i=0; i < numItems; ++i)
			{
				if (EOF == fscanf (fi, "%" NAME_BUFF_SIZE_STRING "s", nameBuff))
					break;
				if ( ! (itemNames[i] = mystrdup (nameBuff)))
					break;
			}
			if (i < numItems)
				break;
		}

		i = j = 0;

		firstColumn = TRUE;
		dist [numItems-1][numItems-2] = -1000;

		while (i < numItems)
		{
			if (firstColumn)
			{
				if (nameOption == NAMES_IN_DATA)
				{
					if ( EOF == fscanf (fi, "%" NAME_BUFF_SIZE_STRING "s", nameBuff))
						break;
					if ( ! (itemNames[i] = mystrdup (nameBuff)))
						break;
				}
				firstColumn = FALSE;
			}

			if (EOF == fscanf (fi, SCANF_FLOAT_FORMAT, & f))
				break;

			switch (matrixOption)
			{
			case SQUARE_MATRIX:
				if (i > j)
					dist[i][j] = f;
				++ j;
				if (j >= numItems)
				{
					++i;
					j = 0;
					firstColumn = TRUE;
				}
				break;
	
			case UPPER_T_MATRIX:
				dist [j][i] = f;	/* transpose */
				++ j;
				if (j >= numItems)
				{
					++i;			
					j = i;			/* new row, start at correct column */
					firstColumn = TRUE;
				};
				break;

			case LOWER_T_MATRIX:	
				dist [i][j] = f;
				++ j;
				if (j > i)			/* order is different; we might not have found the diagonal zero */
				{
					++ i;
					j = 0;
					firstColumn = TRUE;
				}
				break;
			}
		}
		if (dist [numItems-1][numItems-2] == -1000)	/* break if not all items read */
			break;

		myfclose (fi);

		if (saveCopy)	/* clone the distances, if necessary */
		{
			if ( ! (distClone = (Float**) allocLowerTriangle (numItems, sizeof (Float))))
			{
				puts ("No room to save a copy of distances; no comparison will be done.");
				compareDist = FALSE;
			}
			else
			{
				for (i=1; i < numItems; ++i)
					memcpy (distClone[i], dist[i], i * sizeof (Float));
			}
		}
		return TRUE;	/* all done */

	} while (FALSE);

	if (fi)
		fclose (fi);

	printf ("Ran out of data reading file.\n");

	return FALSE;	/* say we crapped out */
}


void fprintNodename (FILE *f, int i, char** nodeNames, int isLeaf)
{
	/* print a suitable name for the node */

	if (i >= 0)
	{
		if (nodeNames)
			fprintf (f, "%s", nodeNames [i]);
		else
			fprintf (f, "Item_%-5d", i);
	 }
	 else if (! isLeaf)
	 {
		fprintf (f, "%d", - i );
    }
    else
    {
      fprintf (f, "Group_%d", 1 + groupNums[i + numItems] );
    }
}


#define FLOAT_MAX_VALUE 1e20

int doSAHN (int iopt, int preserveOrder, int* ia, int* ib, Float* crit, Float** diss, char** nodeNames, int showProgress)
{
	/*  quick translation of Fionn Murtagh's cluster algorithm from Statlib */
	/*  but with his one-dimensional upper-triangular diss() array replaced */
	/*  with the lower triangular diss[][] which includes the zero diagonal */
	
	/*  also fixed a bug affecting the two centroid methods */
/*
c++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++c
c                                                            c
c  hierarchical clustering using (user-specified) criterion. c
c                                                            c
c  parameters:                                               c
c                                                            c
c  diss[][]          dissimilarities in lower half diagonal  c
c                    storage; len = n.(n+1)/2,               c
c  iopt              clustering criterion to be used,        c
c  ia, ib, crit      history of agglomerations; dimensions   c
c                    n, first n-1 locations only used,       c
c  membr, nn, disnn  vectors of length n, used to store      c 
c                    cluster cardinalities, current nearest  c
c                    neighbour, and the dissimilarity assoc. c
c                    with the latter.                        c
c  flag              boolean indicator of agglomerable obj./ c
c                    clusters.                               c
c                                                            c
c  f. murtagh, esa/eso/stecf, garching, february 1986.       c
c                                                            c
c------------------------------------------------------------c
*/

/*
c
c  initializations
c
 */
	int i, j, k, n, ncl, im, jm, i2, j2, jj, tot;
	Float dmin, xx, dis1, dis2, newdis;
	Float e;

	int*	nodeNum;	/* number of the node occupying this slot: 0 means none, negative means interior */
	int*	membr;		/* size of the cluster represented by this node slot */	
	int*	nn;			/* index of closest cluster to this one */
	Float*	disnn;		/* distance to closest cluster */

	e = 0;
	do
	{
		if ( ! (nodeNum = myCalloc (sizeof (int), numItems)))
			break;

		if ( ! (membr = myCalloc (sizeof (int), numItems)))
			break;

		if ( ! (nn = myCalloc (sizeof (int), numItems)))
			break;

		if ( ! (disnn = myCalloc (sizeof (Float), numItems)))
			break;

		n = numItems;


		for (i=0; i < n; ++i)
		{
			membr [i] = 1;
			nodeNum[i] = 1 + i;
		}


		ncl = n;
	/*
	c
	c  carry out an agglomeration - first create list of nearest neighbours
	c
	 */
		im = jm = jj = 0;

		if (preserveOrder)
		{
		/* nearest neighbour has to be before or after! */
			disnn[0] = diss[1][0];
			nn[0] = 1;
			disnn[n-1] = diss[n-1][n-2];
			nn[n-1] = n-2;

			for (i=1; i < n - 1; ++i)
				if (diss[i][i-1] < diss[i+1][i])
				{
					disnn[i] = diss[i][i-1];
					nn[i] = i-1;
				}
				else
				{
					disnn[i] = diss[i+1][i];
					nn[i] = i+1;
				}
		}
		else
		{
			for (i=1; i < n; ++i)
			{
				dmin = FLOAT_MAX_VALUE;
				for (j=0; j < i; ++j)
				{
					if (diss[i][j] < dmin)
					{
						dmin = diss[i][j];
						jm = j;
					}
				}
				nn [i] = jm;
				disnn [i] = dmin;
			}
		}

		if (showProgress)
			printf (" Step  Node A      Node B        Distance\n"
					"===========================================\n"
					);

	
		do		/*  main clustering loop */
		{

/*
c     next, determine least diss. using list of nearest neighbours
 */
			dmin = FLOAT_MAX_VALUE;

			for (i=1; i < n; ++i)
			{
				if ((nodeNum[i]) && disnn[i] < dmin)
				{
					dmin = disnn[i];
					im = i;
					jm = nn[i];
				}
			}

/*
c
c  this allows an agglomeration to be carried out.
c
*/
/*	it is always the case that im > jm, so the following

			i2 = Math.max (im, jm);
			j2 = Math.min (im, jm);

	is replaced by:
*/
			i2 = im;
			j2 = jm;

			ia [n-ncl] = nodeNum [j2] - (nodeNum[j2] > 0 ? 1 : 0);		/*  interior nodes are flagged as negative */
			ib [n-ncl] = nodeNum [i2] - (nodeNum[i2] > 0 ? 1 : 0);
			if (iopt == WARDS_MINIMUM_VARIANCE)
			{
				e += dmin / 2;
				crit [n-ncl] = e;
			}
			else
			{
				crit [n-ncl] = dmin;
			}
			nodeNum [i2] = - (n - ncl + 1);		/*  assign negative interior node number */

/*
c
c  update dissimilarities from new cluster.
c
	combine elements i2 and j2 into a new one at index i2, and mark j2 inactive

*/
			nodeNum [j2] = 0;	/* mark node as merged */

			dmin = FLOAT_MAX_VALUE;

			newdis = 0.0;

			for (k=0; k < n; ++k)
			{
				if ((nodeNum[k]) && k != i2)
				{
					dis1 = (i2 > k) ? diss[i2][k] : diss[k][i2];
					dis2 = (j2 > k) ? diss[j2][k] : diss[k][j2];
					xx = (i2 > j2) ? diss[i2][j2] : diss[j2][i2];
					switch (iopt)
					{

					case SINGLE_LINKAGE:
						newdis = MIN (dis1, dis2);
						break;

					case COMPLETE_LINKAGE:
						newdis = MAX (dis1, dis2);
						break;

					case UNWEIGHTED_ARITHMETIC_AVERAGE:
						newdis = (membr[i2] * dis1 + membr[j2] * dis2) / (membr[i2] + membr[j2]);
						break;

					case WEIGHTED_ARITHMETIC_AVERAGE:
						newdis = (dis1 + dis2) / 2;
						break;

					case UNWEIGHTED_CENTROID:
						tot = membr[i2] + membr[j2];
						newdis = (membr[i2]*dis1 + membr[j2]*dis2 - membr[i2] * membr[j2] * xx / tot) / tot;
						break;

					case WEIGHTED_CENTROID:
						newdis = (dis1 + dis2) / 2 - xx / 4;
						break;

					case WARDS_MINIMUM_VARIANCE:
						newdis = ((membr[i2] + membr[k]) * dis1 + (membr[j2] + membr[k]) * dis2 - membr[k] * xx) / (membr[i2] + membr[j2] + membr[k]);
						break;
					}

					if (preserveOrder)
					{
						if (i2 > k)
							diss[i2][k] = newdis;
						else
							diss[k][i2] = newdis;
					}						
					else
					{

						if (i2 > k)
						{
							diss[i2][k] = newdis;
							if (newdis < dmin)		/*  find nearest neighbour of i2 */
							{
								dmin = newdis;
								jj = k;
							}
						}
						else
						{
							diss[k][i2] = newdis;

	/*  FIX: the rest of the else clause is a fix by JB to ensure */
	/*  correct nearest neighbours are found when a non-monotone */
	/*  clustering method (e.g. the centroid methods) are used */

							if (newdis < disnn[k])		/*  find nearest neighbour of i2 */
							{
								disnn[k] = newdis;
								nn[k] = i2;
							}
						}
					}
				}
			}	

			membr[i2] += membr[j2];		/*  add in element count from j2 */
			if (!preserveOrder)
			{
				disnn[i2] = dmin;
				nn[i2] = jj;
	/*
	c
	c  update list of nns insofar as this is required.
	c
	*/
				for (i=1; i < n; ++i)
				{
					if ((nodeNum[i]) && (nn[i] == i2 || nn[i] == j2))
					{
						dmin = FLOAT_MAX_VALUE;
						for (j=0; j < i; ++j)
						{
							if ((nodeNum[j]) && i != j && diss[i][j] < dmin)
							{
								dmin = diss[i][j];
								jj = j;
							}
						}
						nn[i] = jj;
						disnn[i] = dmin;
					}
				}
			}
		
	
			if (preserveOrder && (ncl > 2))	/* completely update list of nearest neighbours */
			{
				for (i=0; i < n; ++i)
					if (nodeNum[i])
						break;

				k = i;

				disnn[k] = FLOAT_MAX_VALUE;

				for (;;)
				{
					/* find next adjacent node, if there is one */
					for (j=k+1; j < n && !nodeNum[j]; ++j)
						/* null loop body*/;

					if (j == n)
						break;

					if (diss[j][k] < disnn[k])
					{
						disnn[k] = diss[j][k];
						nn[k] = j;
					}
					disnn[j] = diss[j][k];
					nn[j] = k;
					k = j;
				}
			}

			if (showProgress)
			{
				printf ("%5d\t", n - ncl + 1);
				fprintNodename (stdout, clustNode1 [n - ncl], nodeNames, FALSE);
				putchar ('\t');
				fprintNodename (stdout, clustNode2 [n - ncl], nodeNames, FALSE);
				printf ("\t%.5f\n", mergeDist1 [n - ncl]);
			}

			--ncl;		/* Progress! one less cluster to go */
/*
	keep agglomerating until finished
*/
		} while (ncl > 1);

		free (nodeNum);
		free (membr);
		free (nn);
		free (disnn);

		return TRUE;
		
	} while (FALSE);

	return FALSE;		/* report we crapped out */
}

int doNeighbour (int* ia, int* ib, Float* length1, Float* length2, Float** dist, char** nodeNames, int showProgress)
{
	/*
		 Saitou and Nei (1987) neighbour joining method
		 The algorithm is O(n^3), at least in the revision
		 presented by Studier and Keppler (1988)
		
		 (NEIGHBOR in the Phylip package is actually implemented
		 in a way that requires time o(N^4))
		
		Recalculation of r[] is optimized by using the following
		shortcuts, which come from the definitions
		of r[] and the new node distances.

		Suppose nodes i and j have just merged to form node u, and let
		k be any other node.  There are N nodes, counting
		i and j but not u.

		Then:

		r[u] = (r[i] + r[j] - N * d(i,j)) / 2

		and

		r'[k] = r[k] - d(i,k) - d(j,k) + d(k,u)
		
		where r'[] is the set of new divergences for nodes other than
		i, j, and u

		These formulas don't improve the order, because we
		still search over all j < i < N for the minimum
		value of Mij, but the constant factor multiplying N^3 is reduced.

		I wanted the output tree to be exactly
		the same as that from Phylip's NEIGHBOR for easy comparison,
		but alas that is not always possible, as the following
		argument shows:

			Let Mij = (N-2) i_j - (Ri + Rj)
			be the rate-corrected distance, where i_j means
			the distance between nodes i and j.

			When there are 4 nodes left, we have

			Mab = 2 a_b - (a_b + a_c + a_d + a_b + b_c + b_d)
				= - (a_c + a_d + b_c + b_d)
				= Mcd
				regardless of which distinct nodes are labelled a, b, c, d

			This means the pair of nodes chosen for merging is not well-defined
			by the algorithm (ie. if Mab is the smallest, so is Mcd).

			In practice, the computation of Mab and Mcd can suffer rounding
			error, and so different implementations might not yield
			the same choice of pair of nodes to be merged.  In fact,
			changing from double to float as the real variable type
			in this program or in Phylip's NEIGHBOR.C can change the
			tree produced.

			Luckily, the choice of one pair of nodes or its complement
			doesn't affect the unrooted tree; when plotted (with DRAWTREE),
			the same tree is obtained.  Only the order of the last three
			steps in the tree building process is affected, not the
			branch lengths or topology.

  */

	Float* r;
	int* nodeNum;					/* number of the node occupying this slot: 0 means none, negative means interior */
	int i, j, k, u=0, N, iMin, jMin;
	Float duk, dik, djk, dij;
	Float M, MMin;

	do
	{
		if ( ! (r = (Float*) myCalloc (sizeof (Float), numItems)))
			break;

		if ( ! (nodeNum = (int*) myCalloc (sizeof (int), numItems)))
			break;

	/* compute divergence, r, the sum of distances to all other nodes */

		for (i=0; i < numItems; ++i)
		{
			nodeNum[i] = i+1;			/*  initial node numbers  */
			for (j=0; j < i; ++j)
			{	
				r[i] += dist[i][j];		/*  add the distance to r for */
				r[j] += dist[i][j];		/*  each node involved */
			}
		}

		if (showProgress)
			printf (" Step  Node A       Length    Node B       Length\n"
					"==================================================\n"
					);


		N = numItems;
		iMin = -1;
		jMin = -1;

		while (N > 2)		/*  while there are at least three nodes left */
		{
		/*  find the smallest value of  */
		/*  Mij := (N-2) * dist[i][j] - (r[i]+ r[j]) */

			MMin = FLOAT_MAX_VALUE;

			for (i=1; i < numItems; ++i)
			{
				if ((nodeNum [i] == 0))			/*  skip if node has been merged */
					continue;

				for (j=0; j < i; ++j)
				{	
					if ((nodeNum [j] == 0))		/*  skip if node has been merged */
						continue;

					M = (N-2) * dist[i][j] - (r[i]+r[j]);

					if (M < MMin)
					{
						MMin = M;
						iMin = i;
						jMin = j;
					}
				}
			}
		/*  note which nodes are being merged */
		/*  (this ordering yields the same (planar) tree as Phylip's NEIGHBOR program) */

			ia [numItems - N] = nodeNum [jMin] - (nodeNum[jMin] > 0 ? 1 : 0);		
			ib [numItems - N] = nodeNum [iMin] - (nodeNum[iMin] > 0 ? 1 : 0);

		/*  replace node j with new node u, the merge of i and j */
		
			u = jMin;		/*  for convenience, u is the index for the merged node */

			nodeNum [u] = (-1) - (numItems - N);			/*  assign negative interior node number */
			nodeNum [iMin] = 0;								/*  mark as merged */

		/*  record distances from merged nodes to their merger */
		/*  (making sure to match the order of nodes from above) */

			dij = dist[iMin][jMin];
			length2 [numItems - N] = (dij + (r[iMin] - r[jMin])/(N-2)) / 2;
			length1 [numItems - N] = dij - length2[numItems - N];

/*
	I thought perhaps NEIGHBOR did this moving of negative lengths
	to the adjacent node.  Doesn't seem like it does, so I've
	left this out.

			if (length1 [numItems - N] < 0)
			{
				length1 [numItems - N] = 0;
				length2 [numItems - N] = dij;
			}
			else if (length2 [numItems - N] < 0)
			{
				length2 [numItems - N] = 0;
				length1 [numItems - N] = dij;
			}
*/
  
		/*  compute divergence of new node */

			r[u] = (r[iMin] + r[jMin] - dij * N) / 2;
		
		/*  compute distances to new nodes,  */
		/* 		d(u,k) = (d(iMin,k) + d(jMin,k) - d(iMin,jMin)) / 2 */

		/*  and recompute divergences of other nodes */
		/* 		r'(k) = r(k) + d(u,k) - d(iMin,k) - d(jMin,k) */


			for (k=0; k < numItems; ++k)
			{
				if ((nodeNum [k] == 0) || k == u)	/*  don't recompute for merged nodes or new node */
					continue;

			/*  compute the distance d(u,k): */
			
				/*  get d(iMin, k) */

				if (iMin > k)						
					dik = dist[iMin][k];
				else
					dik = dist[k][iMin];

				/*   get d(jMin, k) */

				if (jMin > k)
					djk = dist[jMin][k];
				else
					djk = dist[k][jMin];

				/*  subtract dij and divide by 2 to get d(u,k) */

				duk = (dik + djk - dij) / 2;

				r[k] += duk - dik - djk;

				/*  we no longer need the old value of d(jMin,k) */
				/*  so overwrite it with the new value d(u,k) */

				if (u > k)
					dist[u][k] = duk;
				else
					dist[k][u] = duk;
			}

			if (showProgress)
			{
				printf ("%5d\t", numItems - N + 1);
				fprintNodename (stdout, clustNode1 [numItems - N], nodeNames, FALSE);
				printf ("\t%.5f\t", mergeDist1 [numItems - N]);
				fprintNodename (stdout, clustNode2 [numItems - N], nodeNames, FALSE);
				printf ("\t%.5f\n", mergeDist2 [numItems - N]);
			}
		
			--N;		/*  reduce number of nodes left */
		}			/*  keep merging until only two nodes are left */

		/*  find the remaining node: */

		for (i=0; i < numItems; ++i)
			if (i != u && ! (nodeNum [i] == 0))
				break;
	
		/* connect the remaining node to the 'root' */

		ia [numItems - N] = nodeNum [u] - (nodeNum[u] > 0 ? 1 : 0);		
		ib [numItems - N] = nodeNum [i] - (nodeNum[i] > 0 ? 1 : 0);
		
		if (i > u)
			length1[numItems - N] = length2[numItems - N] = dist[i][u];
		else
			length1[numItems - N] = length2[numItems - N] = dist[u][i];


		free (r);
		free (nodeNum);
		return TRUE;

	} while (FALSE);

	return FALSE;	/* something crapped out */
}

/*
	ELTree:  edge-labelled tree for use by QCLUST
		nodes are labelled by integers, edges by Floats
*/


/* characters used by Phylip DRAWTREE: */

#define OPEN_SUBTREE	'('
#define CLOSE_SUBTREE	')'
#define DELIM_LABEL		':'
#define DELIM_SUBTREE	','


typedef struct ELTreeStruct
{
	int numChildren;		/*  number of child trees */
	int numNodes;			/*  total nodes in tree rooted here */
	Float depth;			/*  depth of subtree: maximum distance of any descendent from the root */
	int	tmp;				/*  temporary storage for, say, sorting the tree */

	int node;				/*  label for this node */
							
	int Tlength;			/*   number of elements in arrays T and L, at least numChildren, possibly more */
	
	struct ELTreeStruct** T;/*  the children of this node */
	
	Float* L;				/*  "labels" for the edges to the child trees */

} * ELTree;
	
ELTree newELTree (int childCapacity)
{
		/*  construct an unfilled non-leaf node */

	ELTree this = (ELTree) myCalloc (sizeof (struct ELTreeStruct), 1);
	if (!this)
		return NULL;

	this->numChildren = 0;
	this->node = 0;
	this->numNodes = 1;
	if (childCapacity > 0)
	{
		if (! (this->T = (ELTree*) myCalloc (sizeof (ELTree), childCapacity)))
			return NULL;

		if (! (this->L = (Float*) myCalloc (sizeof (Float), childCapacity)))
			return NULL;
	}
	else
	{
		this->T = NULL;
		this->L = NULL;
	}
	this->Tlength = childCapacity;

	return this;
}

ELTree newLeafNode (int node)
{
		/*  construct a leaf node */
	
	ELTree this = (ELTree) myCalloc (sizeof (struct ELTreeStruct), 1);

	if (!this)
		return NULL;

	this->numChildren = 0;
	this->node = node;
	this->T = NULL;
	this->Tlength = 0;
	this->L = NULL;
	this->numNodes = 1;
	this->depth = 0.0;

	return this;
}

int addChild (ELTree p, ELTree cT, Float cL)
{
	/*  add a child node */
	if (p->numChildren == p->Tlength)		/*  expand child array if necessary */
	{
		if (! (p->T = realloc (p->T, 2 * p->numChildren * sizeof (ELTree))))
			return FALSE;
		if (! (p->L = realloc (p->L, 2 * p->numChildren * sizeof (Float))))
			return FALSE;

		p->Tlength *= 2;
	}
	p->T [p->numChildren] = cT;
	p->L [p->numChildren] = cL;
	++ p->numChildren;
	p->numNodes += cT->numNodes;
	return TRUE;
}

void swapChildren (ELTree p, int c1, int c2)
{
	ELTree tmpE = p->T[c1];
	Float tmpf = p->L[c1];

	p->T[c1] = p->T[c2];
	p->T[c2] = tmpE;

	p->L[c1] = p->L[c2];
	p->L[c2] = tmpf;
}

ELTree fromClusterOutput (int numItems, int* node1, int* node2, Float* clusterDistance)		
	
	/*  Build a tree from clustering algorithm output */
	/*  The arrays node1 and node2 are indexed by */
	/*  edge number, and their values are node numbers. */
	/*  Negative node numbers represent interior nodes. */
	/*  Non-negative node numbers represent leaf nodes. */
	
	/*  clusterDistance is an array giving the distance at which cluster */
	/*  combining occurred for each step in the clustering */
	/*  We convert these to distances */
	/*  between clustered nodes and the newly created Inode by */
	/*  remembering the "depth" of each node as we build the tree */

	/* 
	    Note:  interior node numbers must occur in node1[] and node2[]
		in increasing order of magnitude for edge labels to be generated correctly.
		The clustering algorithms ensure this.
	 */

	/*  We construct an ELTree for each interior and leaf, */
	/*  then assign children to parents. */
{
	int numLeaves = numItems;
	int numInodes = numItems - 1;	/* clustering produces a tree with n-1 interior nodes */
	int numNodes = numInodes + numLeaves;
	Float* depth = (Float*) myCalloc (sizeof (Float), numNodes);		/*  elements are initialized to zero */
	ELTree* nodes = (ELTree*) myCalloc (sizeof (ELTree), numNodes);

	int i, j;

	if ((!depth) || (!nodes))
		return NULL;

	/*  allocate and name interior nodes */

	for (i=0; i < numInodes; ++i)
	{
		if (! (nodes[i] = newELTree (2)))
			return NULL;

		nodes[i]->node = -(numInodes - i);
	}


	/*  allocate leaf nodes */

	for (; i < numNodes; ++i)
	{
		if (! (nodes[i] = newELTree (0)))
			return NULL;
		nodes[i]->node = i - numInodes;
	}

	/*  for each interior node, connect its children, */
	/*  labelling the edges with the appropriate distances */

	for (i=0; i < numInodes; ++i)
	{
		j = numInodes - i - 1;
		depth[j] = clusterDistance[i] / 2.0;
		if (!addChild (nodes[j], nodes [numInodes + node1[i]], depth[j] - depth[numInodes + node1[i]]))
			return NULL;
		if (!addChild (nodes[j], nodes [numInodes + node2[i]], depth[j] - depth[numInodes + node2[i]]))
			return NULL;
	}

	/*  the complete ELTree is now rooted at nodes[0] */

	return nodes[0];
}



ELTree fromNeighbourOutput (int numItems, int* node1, int* node2, Float* mergeDist1, Float* mergeDist2)		
	
	/*  Build a tree from neighbour joining algorithm output */
	/*  The arrays node1 and node2 are indexed by */
	/*  edge number, and their values are node numbers. */
	/*  Negative node numbers represent interior nodes. */
	/*  Non-negative node numbers represent leaf nodes. */

	/*  mergeDist1[] and mergeDist2[] are the computed branch */
	/*  lengths between nodes merged at each step, and the new */
	/*  node representing their merger */

	/*  We construct an ELTree for each interior and leaf, */
	/*  then assign children to parents. */
{
	int numLeaves = numItems;
	int numInodes = numItems - 2;	/* neighbour joining produces a tree with n-1 interior nodes */
	int numNodes = numInodes + numLeaves;

	int i, j;

	ELTree* nodes = (ELTree*) myCalloc (sizeof (ELTree), numNodes);

	if (!nodes)
		return NULL;

	/*  allocate and name interior nodes */

	for (i=0; i < numInodes; ++i)
	{
		if (! (nodes[i] = newELTree (2)))
			return NULL;
		nodes[i]->node = -(numInodes - i);
	}


	/*  allocate leaf nodes */

	for (; i < numNodes; ++i)
	{
		if (! (nodes[i] = newELTree (0)))
			return NULL;
		nodes[i]->node = i - numInodes;
	}

	/*  for each interior node but the last, connect its children, */
	/*  labelling the edges with the appropriate distances */

	for (i=0; i < numInodes; ++i)
	{
		j = numInodes - i - 1;
		if (! addChild (nodes[j], nodes [numInodes + node1[i]], mergeDist1[i]))
			return NULL;
		if (! addChild (nodes[j], nodes [numInodes + node2[i]], mergeDist2[i]))
			return NULL;
	}

	/*  the last merge involves adding a third child to the root */

	if (node1[i] != - numInodes)	/* get whichever remaining node is not the root */
		j = node1[i];
	else
		j = node2[i];

	if (! addChild (nodes[0], nodes [j + numInodes], mergeDist1[i]))
		return NULL;

	/*  the complete ELTree is now rooted at nodes[0] */
	
	return nodes[0];
}

void fprintELTree (FILE *f, ELTree p)
{
	/*  dump out the tree in infix order, */
	/*  considering all closed nodes to be leaf nodes */
	/*  this is actually depth-first order, because */
	/*  we don't output labels for interior nodes */

	int i;
   int isLeaf = p->node < 0 && (- p->node) <= numItems - numGroups;

	if (p->numChildren == 0 || isLeaf)
	{
		fprintNodename (f, p->node, itemNames, isLeaf);
	}
	else
	{
		fputc (OPEN_SUBTREE, f);
		for (i=0; i < p->numChildren; ++i)
		{
			if (i > 0)
				fputc (DELIM_SUBTREE, f);

			fprintELTree (f, p->T[i]);
			fputc (DELIM_LABEL, f);
			fprintf (f, "%.5f", p->L[i]);
		}
		fputc (CLOSE_SUBTREE, f);
	}
}

void fprintStems (FILE *f, int* stems, int depth, int connect)
{
	int i;
	for (i=0; i < depth; ++i)
	{
		if (stems[i])
			fputs ("|   ", f);
		else
			fputs ("    ", f);
	}
	if (connect && depth >= 0)
		fputs ("+---", f);
}

void sortByLeaves (ELTree p)
{
/* sort children by increasing order of their smallest leaf node */

	int i, j;

	if (p->numChildren == 0)
	{
		p->tmp = p->node;
	}
	else
	{
		for (i=0; i < p->numChildren; ++i)
			sortByLeaves (p->T[i]);	

		/* 
			each child knows its smallest leaf node
			so bubble sort them
		*/

		for (i=0; i < p->numChildren; ++i)
			for (j=i+1; j < p->numChildren; ++j)
				if (p->T[i]->tmp > p->T[j]->tmp)
					swapChildren (p, i, j);

		/*  now we know our smallest leaf node */

		p->tmp = p->T[0]->tmp;
	}
}	


void recfprintAsPicture (FILE *f, ELTree p, int* stems, int drawDepth, int lastChild)
{
	int i;

   int isLeaf = p->node < 0 && (- p->node) <= numItems - numGroups;
	if (p->numChildren == 0 || isLeaf)
	{
		fprintStems (f, stems, drawDepth - 1, TRUE);
		fprintNodename (f, p->node, itemNames, isLeaf);
		fputc ('\n', f);
		stems[drawDepth-1] = ! lastChild;
	}      
	else
	{
		for (i=0; i < p->numChildren; ++i)
		{
			recfprintAsPicture (f, p->T[i], stems, drawDepth + 1, i == p->numChildren - 1);

			if (i == 0)
				if (drawDepth > 0)
					stems[drawDepth-1] = ! lastChild;

			if (i == (p->numChildren - 2) / 2)
			{
				fprintStems (f, stems, drawDepth - 1, p->numChildren == 2);
				fprintNodename (f, p->node, itemNames, FALSE);
				fputc ('\n', f);
			}
			else if (i != p->numChildren - 1)
			{
				fprintStems (f, stems, drawDepth + 1, FALSE);
				fputc ('\n', f);
			}
		}
		stems[drawDepth] = FALSE;
	}
}

void fprintAsPicture (FILE *f, ELTree p)
{
	int *stems = myCalloc (sizeof (int), p->numNodes);
	sortByLeaves (p);
	recfprintAsPicture (f, p, stems, 0, TRUE);
	free (stems);
}


void fprintLengthTable (FILE *f, ELTree p)
{
	int i;
   int childNode;

   if (p->node < 0 && (- p->node) <= numItems - numGroups)
     return;

	for (i=0; i < p->numChildren; ++i)
	{
		fprintNodename (f, p->node, itemNames, FALSE);
		fputc ('\t', f);
      childNode = p->T[i]->node;
      if ( childNode < 0 && - childNode <= numItems - numGroups)
			fprintNodename (f, childNode, itemNames, TRUE);
      else
			fprintNodename (f, childNode, itemNames, FALSE);
		fprintf (f, "\t%.5f\n", p->L[i]);

		if (p->T[i]->numChildren != 0)
			fprintLengthTable (f, p->T[i]);
	}
}

ELTree buildTree ()
{
	ELTree theTree;

	if (clusterMethod != NEIGHBOUR_JOINING)
	{
		theTree = fromClusterOutput (numItems, clustNode1, clustNode2, mergeDist1);
	}
	else
	{
		theTree = fromNeighbourOutput (numItems, clustNode1, clustNode2, mergeDist1, mergeDist2);
	}
	return theTree;
}

static int* clustNodes[2];
static int groupCounter = 0;

/* a formula to map the interior node numbers (-1, -2, -3, ...)
   to their locations in the clustNode array (0, 1, 2, ...)
   and to map the leaf node numbers (0, 1, 2, ...)
   to their origin 1 values
 */
#define MAP_INT_NODE(n) (- (1 + n))
#define MAP_LEAF_NODE(n) (1 + n)


void fillListWithDescendents (int currIntNode)
{
   /* fill list with the indicies of nodes descending from the current
   	 interior node */

   int n;

     /* loop over both children */
    for (n = 0; n < 2; ++n)
    {
       if (clustNodes[n][currIntNode] >= 0)
       {
     /* if the child is a leaf node, add it to the list */
     		groupNums[clustNodes[n][currIntNode] + numItems] = groupCounter;
       }
  
       else
       {
  		/*
       	otherwise, call this function recursively on the (interior) child.
        */
          fillListWithDescendents (MAP_INT_NODE(clustNodes[n][currIntNode]));
       }
	 }
}


int * buildGroups ()
{
 /* build a grouplist of numGroups groups from cluster output */

   int intNode;
   int lowestIntNode;
   int n;

   clustNodes[0] = clustNode1;
   clustNodes[1] = clustNode2;
   groupCounter = 0;

	if (clusterMethod != NEIGHBOUR_JOINING)
   {
      groupNums = (int *) calloc (2 * numItems, sizeof(*groupNums));
      if (!groupNums)
      	return NULL;

      /*
         To form numGroups groups, we examine the uppermost
         numGroups - 1 nodes of the tree.  Any of their children
         that are not themselves in that portion of the tree are
         the roots of subtrees whose leaf nodes are precisely
         the numGroups groups.
         We special-case numGroups = 1.
         The first node in each group is flagged by negating
         its index in the (origin one) list.
       */

      if (numGroups > 1 )
      {
       	lowestIntNode = numItems - numGroups;
         for (intNode = numItems - 2; intNode >= lowestIntNode; --intNode)
          {
	     /* loop over both children */
           	 for (n = 0; n < 2; ++n)
             {
                if (clustNodes[n][intNode] >= 0)
                  /* the child is a leaf node (and so in a group by itself) */
                  {
                   	groupNums [clustNodes[n][intNode] + numItems] = groupCounter++;
                  }
                else if ( MAP_INT_NODE(clustNodes[n][intNode]) < lowestIntNode)
                  /* the child is an interior node, but outside the
                     upper part of the tree, so its descendent
                     leaves form a group */
                  {
							 groupNums [clustNodes[n][intNode] + numItems] = groupCounter;
      	    		    fillListWithDescendents (MAP_INT_NODE(clustNodes[n] [intNode]));
                      ++groupCounter;
                  }
             }
          }
       }
      else
        {
         fillListWithDescendents (numItems - 2);
		  }
   }
 	else
   {
    	puts ("ERROR:  can't make groups for the (unrooted!) neighbour-joining method.");
   }
	return groupNums;
}

int clusterItems (int showProgress)
{
	int rv;
	
	if (showProgress)
	{
		printf ("\nUsing method: %s\n\n", methodNames [clusterMethod]);
		if (preserveOrder)
			puts ("Preserving order:  only adjacent clusters will be joined.");
	}

	switch (clusterMethod)
	{
	case NEIGHBOUR_JOINING:
		rv = doNeighbour (clustNode1, clustNode2, mergeDist1, mergeDist2, dist, itemNames, showProgress);
		break;

	default:
		rv = doSAHN (clusterMethod, preserveOrder, clustNode1, clustNode2, mergeDist1, dist, itemNames, showProgress);
		break;
	}

	return rv;
}

int distComparison (ELTree t, int order, compInfo ci)
{
	/* 
		Compare the leaf-node to leaf-node distances in the tree with
		those in the input, computing a sum of magnitudes, sum of squares,
		maximum, and indicies of the pair of leaf nodes for which the
		maximum occurs.

		We do this by infix traversal, maintaining a list of distances between
		all nodes previously visited.

		This algorithm would be more elegant if the distance between a node
		and its parent was stored with the node instead of the parent.

	*/
	
	int i, j, childNum, jNum, me;
	Float d;

	/* compute and record distances of each child to all nodes already visited */

	me = order;

	ci->deIndex [order] = t->node;	/* record this node as visited at time 'order' */

	++order;

	for (i = 0; i < t->numChildren; ++i)
	{
		childNum = t->T[i]->node;				/* which node is this child */

		for (j = 0; j < order; ++j)			/* loop over nodes already visited */
		{
			jNum = ci->deIndex [j];				/* node that was visited at time j */

			d = t->L[i];					/* distance from child i to visited node j */
			if (me > j)	
				d += ci->treeDist [me][j];
			else
				d += ci->treeDist [j][me];

			ci->treeDist [order][j] = d;			/* record in big array */
			
			/* 
				If child and other node are both leaf nodes, then
				get difference between tree distance and input distance
			 */
		
			if (childNum >= 0 && jNum >= 0)
			{
				if (childNum > jNum)
					d -= ci->inputDist [childNum] [jNum];
				else
					d -= ci->inputDist [jNum] [childNum];

				/* record absolute value of distance */
				
				if (d < 0)
					d = - d;

				ci->sumDiffs += d;
				ci->sumSqDiffs += d * d; 

				/* see if this is the biggest difference so far */

				if (d > ci->maxDiff)
				{
					ci->maxDiff = d;
					ci->whichMax1 = jNum;
					ci->whichMax2 = childNum;
				}
			}
		}		/* continue getting distances between this child and nodes already visited */

		/* invoke this recursively on the child*/

		order = distComparison (t->T[i], order, ci);
			
	}	/* continue getting distance for other children */

	return order;
}

void freeIntermediates ()
{
	/* free up big storage we might need */

	freeMatrix ((void**) dist, numItems);
	dist = NULL;
}
	

compInfo doComparison (ELTree t)
{
	int N;
	compInfo ci;

	do
	{
		freeIntermediates ();

		if (! distClone)
			break;

		if (! (ci = myCalloc (1, sizeof (struct compInfoStruct))))
			break;

		ci->inputDist = distClone;

		N = t->numNodes;	/* total nodes in tree */

		if (! (ci->treeDist = (Float**) allocLowerTriangle (N, sizeof (Float))))
			break;

		if (! (ci->deIndex = myCalloc (N, sizeof (int))))
			break;

		ci->sumDiffs = 0;
		ci->sumSqDiffs = 0;
		ci->maxDiff = 0;
		ci->whichMax1 = ci->whichMax2 = -1;

		distComparison (t, 0, ci);
/*		if (distComparison (t, 0, ci) >= N-1) */
			return ci;

	} while (FALSE);

	return NULL;
}

void outputResults (FILE *f, FILE *tf, ELTree p, compInfo ci)
{
	int i;

	double numDis = numItems * (double) (numItems - 1) / 2;
	fprintf (f, VERSION_STRING "\n");
	fprintf (f, "There are %d items.\n", numItems);
	fprintf (f, "Using the %s method...\n", methodNames [clusterMethod]);

	if (ci)
	{
		fprintf (f, "\nThe Difference between Tree Distance and Input Distance has:\n");

		fprintf (f, "    Sum (of magnitudes): %f\n", ci->sumDiffs);
		fprintf (f, "    Sum of squares:      %f\n", ci->sumSqDiffs);

		fprintf (f, "    Root Mean Square:    %f\n", sqrt(ci->sumSqDiffs / numDis));
		fprintf (f, "    Maximum magnitude:   %f\n", ci->maxDiff);
		if (ci->whichMax1 >= 0 && ci->whichMax2 >= 0)
			fprintf (f, "       which occurs between %s and %s\n\n", itemNames [ci->whichMax1], itemNames [ci->whichMax2]);
	}

	if (numGroups > 1)
   {
   	fprintf (f, "\nTree Picture (Topology only - lengths not represented)\n\n");
   	
   	fprintAsPicture (f, p);
   
   	fprintf (f, "\nNode A    Node B     Length"
   				"\n========================================\n");
   	
   	fprintLengthTable (f, p);
   }

	fprintELTree (tf, p);

	fputc (';', tf);		/* DRAWTREE needs this, although it's syntactically redundant */


   if (numGroups < numItems)
   {
    	fprintf (f, "\n Group  Item"
                  "\n=============================\n" );

    	for (i = 0; i < numItems; ++i)
          fprintf (f, "%d\t%s\n", 1 + groupNums[numItems + i], itemNames[i] );
   }

}

int main (int argc, char** argv)
{
	ELTree theTree;
	compInfo ci = NULL;

#ifdef DO_TIMING
	time_t startTime, endTime;
#endif /* DO_TIMING */

	do
	{
		if (! getOptions (argc, argv))
			break;

#ifdef DO_TIMING
		time (&startTime);
#endif /* DO_TIMING */

		if (showProgress)
		{
			puts ("\n" VERSION_STRING "\n");
			printf ("Reading input...\n");
		}

		if (! grabData (infileName, compareDist))
		{
			printf ("Encountered a problem with input file %s\n", infileName);
			break;
		}

		if (showProgress)
			printf ("Read %d items.\n"
					"Clustering...\n", numItems );

      if (numGroups > numItems || numGroups < 1)
        numGroups = numItems;

		if (! clusterItems (showProgress))
			break;

		if (! (theTree = buildTree ()))
			break;

      if (numGroups < numItems && ! buildGroups ())
         break;

		if (compareDist && ! (ci = doComparison (theTree)))
			puts ("Unable to do comparison of input and computed distances.");

		outfile = myfopen (outfileName, "w");
		if (outfile == NULL)
			printf ("Couldn't open %s for output.\n", outfileName);

		treefile = myfopen (treefileName, "w");
		if (treefile == NULL)
		{
			printf ("Couldn't open %s for output.\n", treefileName);
		}

		outputResults (outfile, treefile, theTree, ci);

		if (showProgress && outfile != stdout)
		{
			myfclose (outfile);
			printf ("Output written to %s.\n", outfileName);
		}

		if (showProgress && treefile != stdout)
		{
			printf ("Phylip DRAWTREE usable output written to %s.\n", treefileName);
			myfclose (treefile);
		}

#ifdef DO_TIMING
		time (&endTime);
		printf ("Time taken: %f seconds\n", difftime(endTime, startTime));
#endif /* DO_TIMING */

		exit(0);

	} while (FALSE);

	exit(1);
}

