/*
   toccf.c      -   convert a tab-delimited matrix of character values into
   a file in Cornell Condensed Format

   The input is a rectangular matrix of m rows and n columns,
   possibly split among several files, each having m lines.
   Optionally:
   - column names can occur in the first line of each file
   - row names can occur as the first word of each line
   - the output matrix can include appropriate names
   - the output can be in the form of a table
   Each line in a file contains data for all (if 1 file)
   or a portion of (if >1 files) a row in the matrix.
   Each line in a file must contain the same number of
   items.


 */

#define VERSION_STRING "TOCCF V0.1"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#define USE_DOUBLE				/* #define USE_DOUBLE if you want to use doubles */

#ifdef USE_DOUBLE

typedef double Float;
#define SCANF_FLOAT_FORMAT "%lf"

#else

typedef float Float;
#define SCANF_FLOAT_FORMAT "%f"

#endif /* USE_DOUBLE */

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE (-1)
#endif
#ifndef NULL
#define NULL ((void *)0)
#endif

/* two definitions, not side-effect proof! */
#ifndef MIN
#define MIN(A,B) ((A<=B) ? (A) : (B))
#endif
#ifndef MAX
#define MAX(A,B) ((A>=B) ? (A) : (B))
#endif

#define INFILE_NAME	"infile"
#define OUTFILE_NAME "outfile"


/* variables used throughout this program */

/* program options */

static int haveRowNames;		/* row names are present */
static int haveColNames;		/* column names are present */
static int samplesAreRows;		/* true if samples are rows, not columns */
static int numRows;				/* number of rows */
static int numCols;				/* number of columns */
static int numSample;			/* number of items between which distances are computed; either numRows or numCols */
static int numChar;				/* number of attributes; i.e. length of vector representing a row or column */
static int numInfiles;			/* number of input files */
static int tabDelimited;		/* are cells always delimited by tabs, allowing for empty ones? */
static int binominalNames;		/* are character names binominals to be printed as ggg_sss? */
static char **infileName;		/* name(s) of input file(s) */
static char *outfileName;		/* name of output file */
static FILE **infile;			/* pointer(s) to input file(s) */
static FILE *outfile;			/* pointer to output file */
static char **rowNames;			/* names of rows */
static char **colNames;			/* names of columns */
static char **sampleName;		/* names of rows or columns, whichever are the samples */
static char **charName;			/* names of rows or columns, whichever are the characters (species) */
static enum
  {
	OF_TWINSPAN, OF_CANOCO
  }
outputFormat;					/* output format */
static Float **inputMat;		/* input matrix */
static int charValWidth;		/* field width for printing character values */
static int charValDecimal;		/* number of digits after decimal point for printing character values */
static int coupletsPerLine;		/* number of couplets per line in output */

/* miscellaneous */

/* these two must match: */

#define ITEM_BUFF_SIZE 50
#define ITEM_BUFF_SIZE_STRING "50"

char itemBuff[ITEM_BUFF_SIZE + 2];	/* buffer to hold text for an item */

void *
myCalloc (int number, int size)
{
  void *res = NULL;

  if (number && size)			/* don't try to allocate 0 bytes - some systems crap out */
	if (!(res = (void *) calloc (number, size)))
	  printf ("Out of memory asking for %ld bytes.\n", size * (long) number);

  return res;
}

void
freeMatrix (void **t, int num)
{
  /*      free the matrix pointed to by t */

  int i;

  if (t)
	{
	  for (i = 0; i < num; ++i)
		if (t[i])
		  free (t[i]);
	  free (t);
	}
}

void **
allocRectangle (int rows, int cols, int size)
{
/* 
   allocate a lower triangular matrix with room for diagonal elements
   the matrix will have num rows and elements of the given size */

  void **t;
  int i;

  do
	{
	  if (!(t = myCalloc (rows, sizeof (void *))))
		  break;

	  for (i = 0; i < rows; ++i)
		if (!(t[i] = myCalloc (cols, size)))
		  break;

	  if (i == rows)
		return t;

	}
  while (FALSE);

  freeMatrix (t, rows);

  return NULL;
}


char *
mystrdup (char *s)
{
  char *res = (char *) myCalloc (sizeof (char), strlen (s) + 1);
  if (res)
	strcpy (res, s);
  return res;
}

#define NUM_REQUIRED_ARGS 2

void
showUsage (char *name)
{
  printf ("Usage: %s ROWS COLS [options...][inputfile...]", name);
  puts ("   where:");
  puts ("ROWS        is the number of rows in the input matrix");
  puts ("COLS        is the number of columns in the input matrix");
  puts ("inputFile   is a list of 0 or more files to be joined horizontally");
  puts ("   and options are:");
  puts ("-h          print this summary of usage");
  puts ("-nr         names of rows are present as the first word in each line");
  puts ("-nc         names of columns are present as the first line in each file");
  puts ("-o outFile  put the matrix into outfile, instead of printing to the terminal");
  puts ("-r          samples are rows, not columns");
  puts ("-t          items are always separated by TABs; a missing item is treated as 0");
  puts ("-b          binominal character names; print as ggg_sss");
  puts ("-fc         format output for reading into Canoco (this is the default)");
  puts ("-ft         format output for reading into Twinspan");
  puts ("-sFMT       use FMT as the Fortran-style format for couplets");
  puts ("            For example, -s5F6.2 means use 5 couplets per line");
  puts ("            and print each character value so that it uses a");
  puts ("            total of 6 spaces, with two digits after the decimal.");
  puts ("            The default format is -s5F5.1");

  puts ("The default behaviour is to read a matrix from standard input,");
  puts ("and print the data in Cornell Condensed Format.");
  puts ("If more than one input file is listed, they are pasted together,");
  puts ("so that the nth row of the input matrix is the nth line of");
  puts ("file1, followed by the nth line of file2, and so on.");

  puts ("\nThis is " VERSION_STRING " by John Brzustowski (jbrzusto@gpu.srv.ualberta.ca)");
  puts ("You are free to use, copy, and modify it without fee.");
}

int
getOptions (int argc, char **argv)
{
  int i;
  char c;
  int crapOut = FALSE;
  int numReqGotten = 0;			/* number of required arguments obtained so far */
  char *restOfArg;

  /* set default options */

  haveRowNames = haveColNames = FALSE;
  numRows = numCols = 0;
  samplesAreRows = FALSE;
  numInfiles = 0;
  tabDelimited = FALSE;
  binominalNames = FALSE;
  outputFormat = OF_CANOCO;
  coupletsPerLine = 5;
  charValWidth = 5;
  charValDecimal = 1;

  /* create enough room to hold all arguments that might be input file names
     a bit wasteful, but big deal */

  infileName = (char **) myCalloc (argc, sizeof (char *));
  infile = (FILE **) myCalloc (argc, sizeof (FILE *));

  outfileName = "Standard_Output";
  outfile = stdout;

  i = 1;

  while (!crapOut)
	{
	  if (i >= argc)
		{
		  if (numInfiles == 0)	/* if no input files, use stdin */
			{
			  ++numInfiles;
			  infile[0] = stdin;
			  infileName[0] = "Standard_Input";
			}
		  if (samplesAreRows)
			{
			  numSample = numRows;
			  numChar = numCols;
			}
		  else
			{
			  numSample = numCols;
			  numChar = numRows;
			}

		  if (numReqGotten == NUM_REQUIRED_ARGS)
			return TRUE;
		  printf ("%s: missing required arguments. Use the -h option for help.", argv[0]);
		  return FALSE;
		}

	  if (argv[i][0] != '-')	/* not an option, either a filename or a required arg */
		{
		  switch (numReqGotten)
			{
			case 0:
			  if ((1 != sscanf (argv[i], "%d", &numRows))
				  ||
				  (numRows <= 0)
				)
				{
				  printf ("%s: number of rows must be > 0\n", argv[0]);
				  crapOut = TRUE;
				}
			  ++numReqGotten;
			  break;

			case 1:
			  if ((1 != sscanf (argv[i], "%d", &numCols))
				  ||
				  (numCols <= 0)
				)
				{
				  printf ("%s: number of columns must be > 0\n", argv[0]);
				  crapOut = TRUE;
				}
			  ++numReqGotten;
			  break;

			default:
			  infileName[numInfiles] = argv[i];
			  if (!(infile[numInfiles] = fopen (argv[i], "r")))
				{
				  printf ("Couldn't open file %s for input.\n", argv[i]);
				  crapOut = TRUE;
				}
			  ++numInfiles;
			  break;
			}
		}
	  else
		{
		  c = argv[i][1];

		  switch (c)
			{
			  /* first handle options that take no argument */
			case 'r':
			  samplesAreRows = TRUE;
			  break;

			case 't':
			  tabDelimited = TRUE;
			  break;

			case 'b':
			  binominalNames = TRUE;
			  break;

			case 'h':
			  showUsage (argv[0]);
			  crapOut = TRUE;
			  break;

			default:
			  /* deal with options that need an argument */


			  if (c == '\0' || argv[i][2] == '\0')
				{
				  /* in case the user left space after the option letter */
				  if (++i < argc)
					restOfArg = argv[i];
				  else
					restOfArg = "";
				}
			  else
				{
				  restOfArg = 2 + argv[i];
				}
			  switch (c)
				{
				case 'o':
				  if (restOfArg)
					{
					  outfileName = restOfArg;
					  if (!(outfile = fopen (outfileName, "w")))
						{
						  printf ("Couldn't open file %s for output.\n", outfileName);
						  crapOut = TRUE;
						}
					}
				  break;

				case 'n':
				  switch (*restOfArg)
					{
					case 'c':
					case 'C':
					  haveColNames = TRUE;
					  break;
					case 'r':
					case 'R':
					  haveRowNames = TRUE;
					  break;
					default:
					  printf ("%s: argument to -%%c option must be 'r' or 'c'\n", restOfArg);
					  crapOut = TRUE;
					};
				  break;

				case 'f':
				  switch (*restOfArg)
					{
					case 't':
					case 'T':
					  outputFormat = OF_TWINSPAN;
					  break;
					case 'c':
					case 'C':
					  outputFormat = OF_CANOCO;
					  break;
					}
				  break;

				case 's':
				  if (((3 != sscanf (restOfArg, "%dF%d.%d", &coupletsPerLine, &charValWidth, &charValDecimal))
					   && (3 != sscanf (restOfArg, "%df%d.%d", &coupletsPerLine, &charValWidth, &charValDecimal)))
					  || coupletsPerLine < 1 || charValWidth < 1 + charValDecimal)
					{
					  printf ("Couldn't understand the -s%s option.\n", restOfArg);
					  puts ("Specify as -sNFW.D where N is the number of couplets,");
					  puts ("W is the total field width, and D is the number of");
					  puts ("decimal places.  e.g. -s5F6.2 means 5 couplets per line");
					  puts ("with each character value taking 6 spaces, and two digits");
					  puts ("after the decimal point.  W must be at least one more than D.");
					  crapOut = TRUE;
					}
				  break;


				default:
				  printf ("%s: unknown option %s\n", argv[0], argv[i - 1]);
				  puts ("Use the -h option for help.");
				  return FALSE;
				}
			}

		};

	  ++i;						/* next arg */

	};
  return FALSE;					/* shouldn't get here */
}


int
nextItem ()
{
  /* get the next item from the appropriate file 
     This means checking to see whether we've reached the end of line
     on the current file, and if so, moving to the next file
     An awkward function, but prevents having to allocate possibly very
     large line buffers that fgets would require.  We don't know the
     size we'd need, anyway.
   */

  static fileNum = 0;			/* keep track of which file we are currently reading from */
  char c;
  int i;
  static int notUsingTabsMsg = FALSE;

  if (tabDelimited)
	{
	  for (;;)
		{
		  c = fgetc (infile[fileNum]);
		  if (c == EOF)
			{
			  ++fileNum;
			  if (fileNum >= numInfiles)
				return FALSE;	/* done reading all files, no item found */
			  break;
			}
		  else
			{
			  for (i = 0; (isgraph (c) || c == ' ') && i < ITEM_BUFF_SIZE; ++i)
				{
				  itemBuff[i] = c;
				  c = fgetc (infile[fileNum]);
				}
			  itemBuff[i] = '\0';
			  if (!isspace (c))	/* if the extra char is not just the expect TAB or EOL delimiter */
				ungetc (c, infile[fileNum]);	/* then return it */
			  if (c != '\t' && (!notUsingTabsMsg) && i == ITEM_BUFF_SIZE)
				{
				  puts ("Hmmm... this file appears to use spaces, not TABS to separate items, and\n"
						"I'm not smart enough to interpret a file with missing data and no TABS!");
				  notUsingTabsMsg = TRUE;
				  return FALSE;
				}

			  if (c == '\r' || c == '\n')
				{
				  c = fgetc (infile[fileNum]);	/* just in case it's a CR-LF pair, get the next char */
				  if (c != '\n')
					ungetc (c, infile[fileNum]);

				  ++fileNum;	/* end of line on this file, move to the next */
				  if (fileNum >= numInfiles)
					fileNum = 0;
				}
			  return TRUE;
			}
		}
	}
  else
	{
	  for (;;)
		{
		  c = fgetc (infile[fileNum]);
		  switch (c)
			{
			case EOF:
			  ++fileNum;
			  if (fileNum >= numInfiles)
				return FALSE;	/* done reading all files, no item found */
			  break;

			case '\r':
			case '\n':
			  c = fgetc (infile[fileNum]);	/* just in case it's a CR-LF pair, get the next char */
			  if (c != '\n')
				ungetc (c, infile[fileNum]);

			  ++fileNum;		/* end of line on this file, move to the next */
			  if (fileNum >= numInfiles)
				fileNum = 0;

			  break;

			default:

			  if (c == '\t' || c == ' ')
				break;
			  for (i = 0; isgraph (c) && i < ITEM_BUFF_SIZE; ++i)
				{
				  itemBuff[i] = c;
				  c = fgetc (infile[fileNum]);
				}

			  ungetc (c, infile[fileNum]);	/* return extra character */

			  itemBuff[i] = '\0';
			  return TRUE;
			}
		}
	}
  return FALSE;					/* shouldn't get here anyway */
}

int
grabData ()						/* any input files are already open */
{
  int i, j;
  Float f;
  char *retPtr;

  i = j = 0;

  do
	{

	  /* allocate space for names, and data */

	  /* allocate space for the input matrix */

	  if (samplesAreRows)
		{
		  if (!(inputMat = (Float **) allocRectangle (numRows, numCols, sizeof (Float))))
			break;
		}
	  else
		{
		  if (!(inputMat = (Float **) allocRectangle (numCols, numRows, sizeof (Float))))
			break;
		}

	  /* allocate a list of names, if there are any */

	  if (haveRowNames && !(rowNames = (char **) myCalloc (numRows, sizeof (char *))))
		  break;

	  if (haveColNames && !(colNames = (char **) myCalloc (numCols, sizeof (char *))))
		  break;

	  /* read data from the files */

	  if (haveColNames)
		{
		  /*

		     BUG FIX:  jmb 17 September '97
		     Throw away the top left corner entry, which should be blank, 
		     if items are tab delimited and row names are present.
		     This entry corresponds to the empty cell in
		     the top left corner of such a matrix.

		   */

		  if (tabDelimited & haveRowNames)
			nextItem ();

		  for (j = 0; j < numCols; ++j)
			{
			  if (!nextItem ())
				{
				  puts ("Couldn't read column name.");
				  goto crapOut;
				}

			  if (strspn (itemBuff, " \t\n") == strlen (itemBuff))
				{
				  fprintf (stderr, "Warning: column %d has no name, I'm calling it 'col%d'.\n", j+1, j+1);
				  sprintf (itemBuff, "col%d", j+1);
				}

              if (!(colNames[j] = mystrdup (itemBuff)))
				{
				  goto crapOut;
				}
			}
		  if (j < numCols)
			break;
		}

	  for (i = 0; i < numRows; ++i)
		{
		  if (haveRowNames)
			{
			  if (!nextItem ())
				{
				  puts ("Couldn't read row name.");
				  goto crapOut;
				}

			  if (strspn (itemBuff, " \t\n") == strlen (itemBuff))
				{
				  fprintf (stderr, "Warning: row %d has no name, I'm calling it 'row%d'.\n", i+1, i+1);
				  sprintf (itemBuff, "row%d", i+1);
				}

			  if (!(rowNames[i] = mystrdup (itemBuff)))
				{
				  goto crapOut;
				}
			}
		  for (j = 0; j < numCols; ++j)
			{
			  if (!nextItem ())
				{
				  puts ("Ran out of data.");
				  goto crapOut;
				}
			  if (*itemBuff == '\0')	/* treat empty cell as zero */
				f = 0;
			  else
				{
				  f = strtod (itemBuff, &retPtr);
				  if (retPtr == itemBuff)
					{
					  printf ("Looking for number but found '%s'.\n", itemBuff);
					  goto crapOut;
					}
				}

			  if (samplesAreRows)
				inputMat[i][j] = f;
			  else
				inputMat[j][i] = f;
			}
		}
	  for (i = 0; i < numInfiles; ++i)
		fclose (infile[i]);

	  if (samplesAreRows)
		{
		  sampleName = rowNames;
		  charName = colNames;
		}
	  else
		{
		  sampleName = colNames;
		  charName = rowNames;
		}

	  return TRUE;

	crapOut:
	  printf ("Error occurred at row %d, column %d.\n", i + 1, j + 1);
	}
  while (FALSE);

  return FALSE;
}



void
outputAsCCF ()
{
  int i, j, l;
  int linePut;
  char nameBuf[100];
  char lineBuf[80];
  char fortranFormatString[66];
  char cFormatString[30];
  char *spName;

  /* create the Fortran format string for sample lines */

  sprintf (fortranFormatString, "(I10,1X,%d(I5,1X,F%d.%d))", coupletsPerLine, charValWidth, charValDecimal);

  /* create the C format string for characterValues */

  sprintf (cFormatString, "%%5d %%%d.%df", charValWidth, charValDecimal);

  switch (outputFormat)
	{
	case OF_CANOCO:

	  /* put first line, giving title to dataset */

	  sprintf (lineBuf, "CANOCO input via TOCCF from:");
	  l = strlen (lineBuf);
	  for (i = 0; i < numInfiles && l + strlen (infileName[i]) < 80; ++i)
		{
		  lineBuf[l] = ' ';
		  strcpy (lineBuf + l + 1, infileName[i]);
		  l += strlen (infileName[i]) + 1;
		}
	  fprintf (outfile, "%s\n", lineBuf);

	  /* put second line, giving format of dataset */

	  fprintf (outfile, "%s\n", fortranFormatString);

	  /* put third line, giving maximum number of couplets per line
	     Note: no newline is output here */

	  fprintf (outfile, "%d", coupletsPerLine);
	  break;

	case OF_TWINSPAN:

	  /* put first line, giving number of samples, species, a title, and a funny line ending */

	  sprintf (lineBuf, "%5d%5dTWINSPAN input via TOCCF from:", numSample, numChar);
	  l = strlen (lineBuf);
	  for (i = 0; i < numInfiles && l + strlen (infileName[i]) < 76; ++i)
		{
		  lineBuf[l] = ' ';
		  strcpy (lineBuf + l + 1, infileName[i]);
		  l += strlen (infileName[i]) + 1;
		}
	  fprintf (outfile, "%-76s** S\n", lineBuf);

	  /*
	     put second line, giving format of dataset and number of couplets per line
	     Note: no newline is output here
	   */

	  fprintf (outfile, "%-65s%5d", fortranFormatString, coupletsPerLine);

	  break;
	}

  /* put lines with data */

  for (i = 0; i < numSample; ++i)
	{
	  linePut = FALSE;			// have output something for this line

	  l = 0;					// number of fields left to print on this line

	  for (j = 0; j < numChar; ++j)
		{
		  if (inputMat[i][j] != 0)
			{
			  if (l == 0)
				{
				  l = coupletsPerLine;
				  fprintf (outfile, "\n%10d ", i + 1);
				}
			  fprintf (outfile, cFormatString, j + 1, inputMat[i][j]);
			  --l;
			  linePut = TRUE;
			}
		}
	  if (!linePut)				/* for all-zero samples, output just the number */
		fprintf (outfile, "\n%10d", i + 1);
	}

  /* output dummy sample to mark end of samples */

  fprintf (outfile, "\n%10d", 0);

  /* output lines with character names */

  l = 0;						// number of fields left to print on this line

  for (i = 0; i < numChar; ++i)
	{
	  if (l == 0)
		{
		  putc ('\n', outfile);
		  l = 10;
		}
	  if (binominalNames)
		{
		  /* grab first part of first word of species name */

		  sprintf (nameBuf, "%-3.3s", charName[i]);

		  /* grab first part of second word of species name */

		  if ((spName = strchr (charName[i], ' ')))
			sprintf (nameBuf + 3, "_%-3.3s ", spName + 1);
		  else
			sprintf (nameBuf + 3, "%-4.4s ", charName[i] + 3);
		}
	  else
		{
		  sprintf (nameBuf, "%-7.7s ", charName[i]);
		}

	  fputs (nameBuf, outfile);

	  --l;
	}

  /* output lines with sample names */

  l = 0;						// number of fields left to print on this line

  for (i = 0; i < numSample; ++i)
	{
	  if (l == 0)
		{
		  putc ('\n', outfile);
		  l = 10;
		}
	  fprintf (outfile, "%8.8s", sampleName[i]);
	  --l;
	}
  putc ('\n', outfile);
  fclose (outfile);

}

void
freeStorage ()
{
  if (haveRowNames)
	freeMatrix ((void **) rowNames, numRows);
  if (haveColNames)
	freeMatrix ((void **) colNames, numCols);
  if (inputMat)
	freeMatrix ((void **) inputMat, numSample);
}

int
main (int argc, char **argv)
{
  do
	{
	  if (!getOptions (argc, argv))
		break;

	  if (!grabData ())
		break;

	  outputAsCCF ();

	  freeStorage ();

	  return FALSE;

	}
  while (FALSE);

  return TRUE;
}
